import os
import csv
import itertools
import arcpy

from lib.geo_methods import *


unit_factor = {
    'ft': 0.3048
    , 'mm': 0.001
    , 'in': 0.0254
    , 'm': 1
    , 'cm': 0.01
}


class wbtError(Exception):
    pass


class Toolbox(object):
    def __init__(self):
        """Define the toolbox (the name of the toolbox is the name of the
        .pyt file)."""
        self.label = "Water Balance Tool: Automatic Run"
        self.alias = "wbt"

        # List of tool classes associated with this toolbox
        self.tools = [Step1_Preparedata, Step2_ProcessData, Step3_Storage, Step4_Derived]


class Step1_Preparedata(object):
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Step 1: Prepare Input Data"
        self.description = ""
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Define parameter definitions"""
        params = []
        params.append(
            arcpy.Parameter(
                displayName='Project Folder'
                , name='output_dir'
                , datatype='DEFolder'
                , parameterType='Required'
                , direction='Input'
            )
        )
        params.append(
            arcpy.Parameter(
                displayName='PRISM Folder'
                , name='prism_folder'
                , datatype='DEFolder'
                , parameterType='Required'
                , direction='Input'
            )
        )
        params.append(
            arcpy.Parameter(
                displayName='Project Area'
                , name='project_area'
                , datatype='DEFeatureClass'
                , parameterType='Required'
                , direction='Input'
            )
        )
        auto_dem = arcpy.Parameter(
            displayName='Download a Digital Elevation Model (DEM) from USGS'
            , name='autodem'
            , datatype='GPBoolean'
            , parameterType='Required'
            , direction='Input'
        )
        auto_dem.value = True
        params.append(auto_dem)

        params.append(
            arcpy.Parameter(
                displayName='Provide your own Digital Elevation Model (DEM)'
                , name='dem'
                , datatype='DERasterDataset'
                , parameterType='Optional'
                , direction='Input'
            )
        )



        units = arcpy.Parameter(
            displayName='DEM Units'
            , name='dem_units'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )
        units.filter.type = "ValueList"
        units.filter.list = sorted(unit_factor.keys())
        units.value = 'm'
        params.append(units)

        autosoil = arcpy.Parameter(
            displayName='Download SOIL AWC from NRCS'
            , name='autosoil'
            , datatype='GPBoolean'
            , parameterType='Required'
            , direction='Input'
        )
        autosoil.value = True
        params.append(autosoil)
        params.append(
            arcpy.Parameter(
                displayName='Provide your own Soil AWC'
                , name='soils'
                , datatype='DERasterDataset'
                , parameterType='Optional'
                , direction='Input'
            )
        )
        sunits = arcpy.Parameter(
            displayName='Soil AWC Units'
            , name='soil_units'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )
        sunits.filter.type = "ValueList"
        sunits.filter.list = sorted(unit_factor.keys())
        sunits.value = 'cm'
        params.append(sunits)
        start = arcpy.Parameter(
                displayName='Start Year'
                , name='start_year'
                , datatype='GPLong'
                , parameterType='Required'
                , direction='Input'
            )
        #start.filter.list = year_filter
        params.append(start)
        end = arcpy.Parameter(
                displayName='End Year'
                , name='end_year'
                , datatype='GPLong'
                , parameterType='Required'
                , direction='Input'
            )
        #end.filter.list = year_filter
        params.append(end)
        params.append(
            arcpy.Parameter(
                displayName='Force Download of new solar data'
                , name='force_download_solar'
                , datatype='GPBoolean'
                , parameterType='Optional'
                , direction='Input'
            )
        )
        params.append(
            arcpy.Parameter(
                displayName='Force Download of new PRISM data'
                , name='force_download_prism'
                , datatype='GPBoolean'
                , parameterType='Optional'
                , direction='Input'
            )
        )
        return params

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
        license = arcpy.CheckExtension('Spatial') == u'Available'
        return license

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""
        params = {p.name:p for p in parameters}

        #if raw_dem exists, autfill it
        if not params['dem'].value and params['output_dir'].altered:
            raw_dem_path = os.path.join(params['output_dir'].valueAsText,'raw_dem.tif')
            if arcpy.Exists(raw_dem_path):
                params['dem'].value = raw_dem_path
                params['autodem'].value = False

        #enables or disables parameters based on download option
        if params['autodem'].altered:
            params['dem'].enabled = params['autodem'].valueAsText == 'false'
            if not params['dem'].enabled:
                params['dem_units'].enabled = False
                params['dem_units'].value = 'm'
            else:
                params['dem_units'].enabled = True
        if params['autosoil'].altered:
            params['soils'].enabled = params['autosoil'].valueAsText == 'false'
            if not params['soils'].enabled:
                params['soil_units'].enabled = False
                params['soil_units'].value = 'cm'
            else:
                params['soil_units'].enabled = True

        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""
        return

    def execute(self, parameters, messages):
        """The source code of the tool."""

        import_toolboxes()
        params = {p.name: p.valueAsText for p in parameters}

        # set to overwrite outputs
        arcpy.env.overwriteOutput = True

        arcpy.SetProgressor('default', 'Starting Step 1')
        # checkout needed extenstions
        progress_message('Checking for Spatial Analyst Extension')
        checkout_status = arcpy.CheckOutExtension('Spatial')
        if not checkout_status == 'CheckedOut':
            arcpy.AddError('Could not enable Spatial Analyst Extension')
            raise wbtError('Spatial Analyst checkout: ' + checkout_status)

        # ensure output dir exists, if not create it
        if os.path.exists(params['output_dir']):
            arcpy.AddWarning("Warning: Project folder already exists. Data will be overwritten")
        else:
            os.mkdir(params['output_dir'])

        # extract dem for project area
        progress_message('Processing DEM')
        dem_outpath = os.path.join(params['output_dir'], 'dem_ex.tif')
        if params['autodem'] == 'true':
            progress_message('Downloading DEM')
            arcpy.FetchData.DownloadNED(params['project_area'], params['output_dir'])
            raw_dem = arcpy.Raster(os.path.join(params['output_dir'],'raw_dem.tif'))
        else:
            raw_dem = arcpy.Raster(params['dem'])
        dem_extracted = arcpy.Raster(
            arcpy.Misc.ExtractAndProjectRaster(raw_dem, params['project_area'], dem_outpath,
                                                      True)) * unit_factor[params['dem_units']]


        # dem_extracted.save(dem_outpath)
        grid = arcpy.sa.Int(dem_extracted * 0 + 1)
        grid.save(os.path.join(params['output_dir'], 'grid.tif'))
        del dem_extracted

        # extract soils
        # arcpy.SetProgressorLabel('Processing Soils')
        progress_message('Processing Soils')
        soils_outpath = os.path.join(params['output_dir'], 'soils.tif')
        if params['autosoil'] == 'true':
            progress_message('Downloading Soil AWC')
            arcpy.FetchData.CreateSoilRaster(params['project_area'], params['output_dir'],dem_outpath)
            awc = arcpy.Raster(os.path.join(params['output_dir'], 'soils_awc.tif'))
        else:
            awc = arcpy.Raster(params['soils'])
        soils_extracted = arcpy.Raster(
            arcpy.Misc.ExtractAndProjectRaster(awc, params['project_area'],
                                                      os.path.splitext(soils_outpath)[0] + '_ex.tif', True)) * unit_factor[params['soil_units']] * 1000
        soils_fixedwater = arcpy.sa.SetNull(soils_extracted < 0.0001, soils_extracted)
        soils_fixedwater.save(os.path.join(params['output_dir'],'soils_fixed.tif'))
        soils_snapped = arcpy.Raster(arcpy.Misc.SnapRaster(soils_fixedwater, grid, soils_outpath))
        arcpy.Delete_management(soils_extracted)
        arcpy.Delete_management(soils_fixedwater)

        # precip
        progress_message('Downloading Precip')
        download_folder = os.path.join(params['prism_folder'], 'prism_raw_ppt')
        if not os.path.isdir(download_folder):
            os.mkdir(download_folder)
        arcpy.FetchData.DownloadMonthlyPrism(download_folder, params['start_year'], params['end_year'], 'ppt',
                                               params['force_download_prism'] == 'true')
        arcpy.FetchData.ExtractPrismData(download_folder, params['output_dir'], params['project_area'], True, grid,
                                           'ppt', params['start_year'], params['end_year'])

        # temperature
        progress_message('Downloading Temp')
        download_folder = os.path.join(params['prism_folder'], 'prism_raw_tmean')
        if not os.path.isdir(download_folder):
            os.mkdir(download_folder)
        arcpy.FetchData.DownloadMonthlyPrism(download_folder, params['start_year'], params['end_year'], 'tmean',
                                               params['force_download_prism'] == 'true')
        arcpy.FetchData.ExtractPrismData(download_folder, params['output_dir'], params['project_area'], True, grid,
                                           'tmean', params['start_year'], params['end_year'])

        # download nsrdb solar
        progress_message('Downloading Solar')
        nsrdb_csv = os.path.join(params['output_dir'], 'nsrdb.csv')
        arcpy.FetchData.DownloadSolarMultipleYears(params['project_area'], int(params['start_year']),
                                                     int(params['end_year']), nsrdb_csv, params['force_download_solar'], 'v3')



        return

class Step2_ProcessData(object):
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Step 2: Calculate Solar and PET"
        self.description = ""
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Define parameter definitions"""
        params = []
        params.append(
            arcpy.Parameter(
                displayName='Project Folder'
                , name='output_dir'
                , datatype='DEFolder'
                , parameterType='Required'
                , direction='Input'
            )
        )
        params.append(
            arcpy.Parameter(
                displayName='Project Area'
                , name='project_area'
                , datatype='DEFeatureClass'
                , parameterType='Required'
                , direction='Input'
            )
        )
        start = arcpy.Parameter(
            displayName='Start Year'
            , name='start_year'
            , datatype='GPLong'
            , parameterType='Required'
            , direction='Input'
        )
        #start.filter.list = year_filter
        params.append(start)
        end = arcpy.Parameter(
            displayName='End Year'
            , name='end_year'
            , datatype='GPLong'
            , parameterType='Required'
            , direction='Input'
        )
        #end.filter.list = year_filter
        params.append(end)
        return params

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
        license = arcpy.CheckExtension('Spatial') == u'Available'
        return license

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""
        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""
        return

    def execute(self, parameters, messages):
        """The source code of the tool."""

        import_toolboxes()
        params = {p.name: p.valueAsText for p in parameters}

        # set to overwrite outputs
        arcpy.env.overwriteOutput = True

        arcpy.SetProgressor('default', 'Starting Step 2')
        dem_outpath = os.path.join(params['output_dir'],'dem_ex.tif')
        nsrdb_csv = os.path.join(params['output_dir'],'nsrdb.csv')

        # run solar analyst
        progress_message('Running Solar Analyst')
        solar_analyst_csv = os.path.join(params['output_dir'], 'solar_analyst.csv')
        arcpy.FetchData.SolarAnalystPointCSV_Multiple(dem_outpath, params['project_area'], solar_analyst_csv)

        progress_message('Finding best D/T values')
        best_dt_csv = os.path.join(params['output_dir'], 'best_dt.csv')
        arcpy.FetchData.FindBestDT(solar_analyst_csv, nsrdb_csv, best_dt_csv)

        progress_message('Runing Solar Analyst Area')
        arcpy.FetchData.SolarAnalystArea(best_dt_csv, dem_outpath, params['output_dir'])

        # Run PET
        progress_message('Calculating PET')
        arcpy.WaterBudget.PET(params['output_dir'], os.path.join(params['output_dir'], 'best_dt.csv'), False, os.path.join(params['output_dir'], 'nsrdb.csv'), int(params['start_year']), int(params['end_year']))

        return

class Step3_Storage(object):
    def __init__(self):
        self.label = "Step 3: Calculate Soil Storage"
        self.description = ""
        self.canRunInBackground = False

    def getParameterInfo(self):
        params = []
        params.append(
            arcpy.Parameter(
                displayName='Project Folder'
                , name='output_dir'
                , datatype='DEFolder'
                , parameterType='Required'
                , direction='Input'
            )
        )
        start = arcpy.Parameter(
            displayName='Start Year'
            , name='start_year'
            , datatype='GPLong'
            , parameterType='Required'
            , direction='Input'
        )
        #start.filter.list = year_filter
        params.append(start)
        end = arcpy.Parameter(
            displayName='End Year'
            , name='end_year'
            , datatype='GPLong'
            , parameterType='Required'
            , direction='Input'
        )
        #end.filter.list = year_filter
        params.append(end)
        skip = arcpy.Parameter(
            displayName='Year 1 January Storage - Wrapping Option'
            , name='skip_wrap'
            , datatype='GPBoolean'
            , parameterType='Optional'
            , direction='Input'
        )
        skip.value = False
        params.append(skip)
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):

        import_toolboxes()

        params = {p.name: p.valueAsText for p in parameters}
        project_folder = params['output_dir']
        start_year = int(params['start_year'])
        end_year = int(params['end_year'])

        # Check all needed data is prepared
        check_missing(project_folder, 'ppt', start_year, end_year, 'Precip', wbtError)
        check_missing(project_folder, 'tmean', start_year, end_year, 'Temperature', wbtError)
        check_missing(project_folder, 'pet', start_year, end_year, 'PET', wbtError)

        # soils
        if not arcpy.Exists(os.path.join(project_folder, 'soils.tif')):
            raise wbtError('Missing Soils')

        # bestdt (and all dt rasters)
        bestdt_months = [(int(row['year']), int(row['month'])) for row in
                         csv.DictReader(open(os.path.join(project_folder, 'best_dt.csv'), 'rb'))]
        if not set(bestdt_months).issuperset(set(itertools.product(range(start_year, end_year + 1), range(1, 13)))):
            raise wbtError('Best DT Csv doesn\'t have all the months')

        # solar area
        bestdt_values = [(float(row['d']), float(row['t'])) for row in
                         csv.DictReader(open(os.path.join(project_folder, 'best_dt.csv'), 'rb'))]
        solar_files = [os.path.join(project_folder, 'solar', 'solar_d%d_t%d.tif' % (d * 10, t * 10)) for d, t in bestdt_values]
        solar_exists = map(arcpy.Exists, solar_files)
        if not all(solar_exists):
            raise wbtError(
                'Missing solar area calculations for %s' % [f[0] for f in zip(solar_files, solar_exists) if not f[1]])

        # Run storage routine
        arcpy.WaterBudget.StorageRoutine(project_folder, start_year, end_year, '#', '#', params['skip_wrap'])
        return

class Step4_Derived(object):
    def __init__(self):
        self.label = "Step 4: Calculate AET, Deficit, and Surplus"
        self.description = ""
        self.canRunInBackground = False

    def getParameterInfo(self):
        params = []
        params.append(
            arcpy.Parameter(
                displayName='Project Folder'
                , name='output_dir'
                , datatype='DEFolder'
                , parameterType='Required'
                , direction='Input'
            )
        )
        start = arcpy.Parameter(
            displayName='Start Year'
            , name='start_year'
            , datatype='GPLong'
            , parameterType='Required'
            , direction='Input'
        )
        #start.filter.list = year_filter
        params.append(start)
        end = arcpy.Parameter(
            displayName='End Year'
            , name='end_year'
            , datatype='GPLong'
            , parameterType='Required'
            , direction='Input'
        )
        #end.filter.list = year_filter
        params.append(end)
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        import_toolboxes()

        params = {p.name: p.valueAsText for p in parameters}
        project_folder = params['output_dir']
        start_year = int(params['start_year'])
        end_year = int(params['end_year'])

        check_missing(project_folder, 'ppt', start_year, end_year, 'Precip', wbtError)
        check_missing(project_folder, 'pet', start_year, end_year, 'PET', wbtError)
        check_missing(project_folder, 'storage', start_year, end_year, 'Soil Storage', wbtError)
        check_missing(project_folder, 'ppe', start_year, end_year, 'P-PE', wbtError)

        arcpy.WaterBudget.StorageDerived(project_folder, start_year, end_year)