import arcpy
import datetime
import zipfile
import os
import csv
import random
import string
import time
import json
import re
import itertools
from lib.geo_methods import *

prism_variables = ['ppt', 'tdmean', 'tmean', 'tmax', 'tmin', 'vpdmax', 'vpdmin']



class DataFetchException(Exception):
    """
    Custom exception that can tell you the tool that caused it as well as include the error message
    """

    def __init__(self, tool, message):
        new_message = tool.label + ' error: ' + message
        arcpy.AddError(new_message)
        super(DataFetchException, self).__init__(new_message)


class Toolbox(object):
    def __init__(self):
        """Define the toolbox (the name of the toolbox is the name of the
        .pyt file)."""
        self.label = "Water Balance Tool: Data Fetching"
        self.alias = "FetchData"

        # List of tool classes associated with this toolbox
        self.tools = [DownloadMonthlyPrism, DownloadSolarMultipleYears, DownloadSolar, DownloadSoilGML, ConvertSoilGMLtoShp, GetSoilAWSTable, CreateSoilRaster, SolarAnalystPointCSV, SolarAnalystPointCSV_Multiple, FindBestDT, DownloadNED, ExtractPrismData, SolarAnalystArea]


class ExtractPrismData(object):
    def __init__(self):
        self.label = "Process: Extract Prism Data by Mask"
        self.description = "Extracts downloaded prism data by mask to an output folder"
        self.canRunInBackground = False

    def getParameterInfo(self):
        params = list()
        params.append(
            arcpy.Parameter(
                displayName='Prism Download Folder'
                , name='download_folder'
                , datatype='DEFolder'
                , parameterType='Required'
                , direction='Input'
            )
        )
        params.append(
            arcpy.Parameter(
                displayName='Output Folder'
                , name='output_dir'
                , datatype='DEFolder'
                , parameterType='Required'
                , direction='Input'
            )
        )
        params.append(
            arcpy.Parameter(
                displayName='Mask'
                , name='mask'
                , datatype='GPFeatureLayer'
                , parameterType='Required'
                , direction='Input'
            )
        )
        extent_only = arcpy.Parameter(
            displayName='Use Extent Only'
            , name='extent_only'
            , datatype='GPBoolean'
            , parameterType='Optional'
            , direction='Input'
        )
        extent_only.value = True
        params.append(extent_only)
        params.append(
            arcpy.Parameter(
                displayName='Reference Raster'
                , name='reference_raster'
                , datatype='DERasterDataset'
                , parameterType='Required'
                , direction='Input'
            )
        )
        metric = arcpy.Parameter(
            displayName='Metric'
            , name='metric'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )
        metric.filter.type = 'ValueList'
        metric.filter.list = prism_variables
        metric.value = 'ppt'
        params.append(metric)
        start_year = arcpy.Parameter(
            displayName='Start Year'
            , name='start_year'
            , datatype='GPLong'
            , parameterType='Required'
            , direction='Input'
        )
        params.append(start_year)

        end_year = arcpy.Parameter(
            displayName='End Year'
            , name='end_year'
            , datatype='GPLong'
            , parameterType='Required'
            , direction='Input'
        )
        params.append(end_year)
        use_spline = arcpy.Parameter(
            displayName='Use Spline Method'
            , name='use_spline'
            , datatype='GPBoolean'
            , parameterType='Optional'
            , direction='Input'
        )
        use_spline.value = False
        params.append(use_spline)
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):

        arcpy.ImportToolbox(os.path.join(os.path.dirname(os.path.abspath(__file__)),'Misc.pyt'))
        params = {p.name: p.valueAsText for p in parameters}
        arcpy.env.workspace = params['output_dir']
        arcpy.env.overwriteOutput = True
        use_spline = params['use_spline'] == 'true'
        check_folder(os.path.join(params['output_dir'], params['metric']))
        # the following seleting of files could probably be more graceful
        files = [path for path in os.listdir(params['download_folder']) if
                 '_' + params['metric'] + '_' in path and path.endswith('.bil') and re.search(
                     '\d{6}', path)]
        files = [x for x in files if int(x.split('_')[-2][0:4]) in range(int(params['start_year']),int(params['end_year'])+1)]
        if not len(files):
            arcpy.AddError('No files found in prism download folder')
            raise DataFetchException(self, 'No files found in prism download folder')

        dates = [(x[0:4], x[4:]) for x in
                 [re.search('\d{6}', f).group(0) for f in files]]
        mask = params['mask']
        mask_describe = arcpy.Describe(mask)
        reference_raster = arcpy.Raster(params['reference_raster'])

        if not use_spline:
            constant_raster = reference_raster * 0 + 1

        arcpy.SetProgressor('step', 'Extracting Prism Files', 0, len(files))
        outfiles = ['%s\%s_%s_%s.tif' % (params['metric'],params['metric'], year, month) for year, month in dates]
        outfiles.extend(
            sorted(['%s\%s_%s_annual.tif' % (params['metric'],params['metric'], year) for year in set([year for year, month in dates])]))
        annual_files = [path for path in os.listdir(params['download_folder']) if
                        '_' + params['metric'] + '_' in path and path.endswith('.bil') and re.search('\d{4}',
                                                                                                     path) and not re.search(
                            '\d{6}', path) and int(path.split('_')[-2][0:4]) in range(int(params['start_year']),int(params['end_year'])+1)]
        files.extend(sorted(annual_files))
        assert len(files) == len(outfiles)
        arcpy.AddMessage(outfiles)
        arcpy.AddMessage(files)
        for out_file, path in zip(outfiles, files):
            arcpy.AddMessage('Reading %s' % path)
            in_rast = arcpy.Raster(os.path.join(params['download_folder'], path))
            assert reference_raster.spatialReference.Name == mask_describe.spatialReference.Name

            if use_spline:
                point_file = os.path.join(params['download_folder'], os.path.basename(out_file))
                if arcpy.Exists(point_file):
                    arcpy.AddWarning('PRISM Point file for %s already exists' % point_file)
                    in_rast_points = point_file
                else:
                    in_rast_points = arcpy.RasterToPoint_conversion(in_rast, point_file)

            if arcpy.Exists(out_file):
                arcpy.AddWarning('Skipping %s cause it already exists' % out_file)
            else:
                arcpy.AddMessage('Extracting %s' % out_file)
                to_delete = []

                if use_spline:
                    in_rast_points_layer = arcpy.MakeFeatureLayer_management(in_rast_points, 'in_rast_points_layer')
                    in_rast_points_layer = arcpy.SelectLayerByLocation_management(in_rast_points_layer,
                                                                                  'WITHIN_A_DISTANCE', mask,
                                                                                  '8 Kilometer', 'NEW_SELECTION')
                    to_delete.append(in_rast_points_layer)

                    in_rast_points_projected = arcpy.Project_management(in_rast_points_layer,
                                                                        'temp_points_projected.shp',
                                                                        mask_describe.spatialReference)
                    to_delete.append(in_rast_points_projected)

                    prism_raster = arcpy.sa.Spline(in_rast_points_projected, 'grid_code',
                                                   min(reference_raster.meanCellHeight, reference_raster.meanCellWidth))
                else:
                    arcpy.env.cellSize = 'MINOF'
                    prism_raster = arcpy.sa.ZonalStatistics(constant_raster, 'VALUE', in_rast, "MEAN")
                    arcpy.env.cellSize = 'MAXOF'

                prism_raster = arcpy.CopyRaster_management(prism_raster, 'prism_rast.tif',
                                                               pixel_type='32_BIT_FLOAT')
                #prism_raster.save('prism_rast.tif')
                to_delete.append(prism_raster)
                random_file = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(5)) + '.tif'
                if not use_spline:
                    prism_clipped = prism_raster

                else:
                    prism_clipped = arcpy.Raster(
                        arcpy.Clip_management(prism_raster, extent_to_envelope(mask_describe.extent), random_file))
                to_delete.append(prism_clipped)

                arcpy.Misc.SnapRaster(prism_clipped, reference_raster, os.path.join(params['output_dir'], out_file))

                for item in to_delete:
                    item_str = str(item)
                    del item
                    arcpy.Delete_management(item_str)

            arcpy.SetProgressorPosition()

        if not use_spline:
            arcpy.Delete_management(constant_raster)
        return


class DownloadMonthlyPrism(object):
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Download: Monthly PRISM Data"
        self.description = "Downloads data from PRISM's ftp server"
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Define parameter definitions"""
        params = list()
        params.append(
            arcpy.Parameter(
                displayName='Download Folder'
                , name='download_folder'
                , datatype='DEFolder'
                , parameterType='Required'
                , direction='Input'
            )
        )
        start_year = arcpy.Parameter(
            displayName='Start Year'
            , name='start_year'
            , datatype='GPLong'
            , parameterType='Required'
            , direction='Input'
        )
        start_year.filter.type = 'ValueList'
        start_year.filter.list = range(1895, datetime.datetime.now().year)
        start_year.value = 1998
        params.append(start_year)

        end_year = arcpy.Parameter(
            displayName='End Year'
            , name='end_year'
            , datatype='GPLong'
            , parameterType='Required'
            , direction='Input'
        )
        end_year.filter.type = 'ValueList'
        end_year.filter.list = range(1895, datetime.datetime.now().year)
        end_year.value = datetime.datetime.now().year - 2
        params.append(end_year)

        metric = arcpy.Parameter(
            displayName='Metric'
            , name='variable'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )
        metric.filter.type = 'ValueList'
        metric.filter.list = prism_variables
        metric.value = 'ppt'
        params.append(metric)

        params.append(
            arcpy.Parameter(
                displayName='Download and Overwrite Existing'
                , name='force_download'
                , datatype='GPBoolean'
                , parameterType='Optional'
                , direction='Input'
            )
        )
        return params

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
        return True

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""
        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""
        return

    def execute(self, parameters, messages):
        """The source code of the tool."""
        params = {p.name: p.valueAsText for p in parameters}
        return self.run(params['download_folder'], params['start_year'], params['end_year'], params['variable'],
                        params['force_download'] == 'true')

    def run(self, download_folder, start_year, end_year, variable, force_download=False):
        """

        :param download_folder: folder to save prism data to
        :param start_year: start year data retreival
        :param end_year: end year for data retreival
        :param variable: the prism code for what variable is desired
        :param force_download: force the download even if file exists already
        :return: download_folder
        """

        # url of ftp site
        url = 'ftp://anonymous:porters1@ohio.edu@prism.nacse.org/monthly/%s/' % variable

        # generate an iterable for years
        years = range(int(start_year), int(end_year) + 1)
        arcpy.SetProgressor('step', 'Downloading data from Prism', 0, len(years), 1)

        # for each year
        for y in years:
            progress_message('Downloading %d' % y)
            filename = 'PRISM_%s_stable_4kmM3_%s_all_bil.zip' % (variable, y)

            if variable == 'tdmean':
                filename = filename.replace('4kmM3', '4kmM1')
            if variable == 'tmean':
                filename = filename.replace('4kmM3', '4kmM2')
            out_path = os.path.join(download_folder, filename)

            # if the file does not exist download it, else add awarning message to indicaet it was skipped
            if not os.path.exists(out_path) or force_download:
                attempts = 0
                failed = True
                while failed and attempts <= 10:
                    url_get = url + str(y) + '/' + filename
                    try:
                        urllib.urlretrieve(url_get, out_path)
                        failed = False
                    except IOError:
                        failed = True
                        attempts += 1
                    except urllib.error.HTTPError as e:
                        arcpy.AddWarning(
                            'Failed to retrieve %s with code %s: %s' % (url_get, e.code, http_error[e.code]))
                        failed = True
                        attempts += 1

                # if the download failed raise an exception
                if failed:
                    if os.path.exists(out_path): os.remove(out_path)
                    DataFetchException(self, 'Failed to download prism file %s' % url + str(y) + '/' + filename)
            else:
                arcpy.AddWarning(str(y) + ' already exists, not downloading again')

            progress_message('Unzipping %d' % y)

            # unzip the file
            zip_ref = zipfile.ZipFile(out_path, 'r')
            zip_ref.extractall(download_folder)
            zip_ref.close()

            arcpy.SetProgressorPosition()

        return download_folder


class DownloadSolarMultipleYears(object):
    def __init__(self):
        self.label = "Download: NSRDB Data for Multiple Years"
        self.description = ""
        self.canRunInBackground = False

    def getParameterInfo(self):
        params = list()
        params.append(
            arcpy.Parameter(
                displayName='Project Area'
                , name='area'
                , datatype='GPFeatureLayer'
                , parameterType='Required'
                , direction='Input'
            )
        )
        start = arcpy.Parameter(
            displayName='Start Year'
            , name='start_year'
            , datatype='GPLong'
            , parameterType='Required'
            , direction='Input'
        )
        #start.filter.list = year_filter
        params.append(start)
        end = arcpy.Parameter(
            displayName='End Year'
            , name='end_year'
            , datatype='GPLong'
            , parameterType='Required'
            , direction='Input'
        )
        #end.filter.list = year_filter
        params.append(end)
        output =  arcpy.Parameter(
                displayName='Output CSV File'
                , name='outfile'
                , datatype='DEFile'
                , parameterType='Required'
                , direction='Output'
        )
        output.filter.list = ['csv']
        params.append(output)
        params.append(
            arcpy.Parameter(
                displayName='Download and Overwrite Existing'
                , name='force_download'
                , datatype='GPBoolean'
                , parameterType='Optional'
                , direction='Input'
            )
        )
        versionparam = arcpy.Parameter(
            displayName='API Version to use'
            , name='version'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Output'
        )
        versionparam.filter.list = ['v2', 'v3']
        versionparam.value = 'v3'
        params.append(versionparam)
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        params = {p.name: p.valueAsText for p in parameters}
        self.run(params['start_year'], params['end_year'], params['area'], params['outfile'],
                 params['force_download'] == 'true', params['version'])
        return

    @staticmethod
    def run(start_year, end_year, project_area, outfile, force_download=False, api_version='v3'):
        # check for existing data
        years = range(int(start_year), int(end_year) + 1)
        download = True
        if os.path.exists(outfile) and not force_download:
            existing_years = [row['year'] for row in csv.DictReader(open(outfile, 'rb'))]
            if set(existing_years) == set(years):
                arcpy.AddWarning('NSRDB data for all desired years already exists, using existing')
                download = False
            else:
                years = set(existing_years) - set(years)
                arcpy.AddWarning('Somedata already exists, only downloading for years:' % years)

        if download or force_download:
            # create empty list to store results
            arcpy.SetProgressor('step', 'Downloading Solar Data', 0, len(years), 1)

            # for each year download data and extend list (returns a list(dict)

            data = []
            for y in years:
                arcpy.SetProgressorLabel('Downloading Solar Data %d' % y)
                data.extend(DownloadSolar.run(y, project_area, api_version))
                arcpy.SetProgressorPosition()

            # save the output to the outfile
            arcpy.SetProgressorLabel('Writing solar data to outfile')
            with open(outfile, 'wb') as outfile:
                csv_w = csv.DictWriter(outfile, fieldnames=[d.strip() for d in data[0].keys()])
                csv_w.writeheader()
                csv_w.writerows([{k.strip():v.strip() for k,v in d.iteritems()} for d in data])

        return


class DownloadSolar(object):
    def __init__(self):
        self.label = "Download: NSRDB Data"
        self.description = ""
        self.canRunInBackground = False

    def getParameterInfo(self):
        params = list()
        params.append(
            arcpy.Parameter(
                displayName='Project Area'
                , name='area'
                , datatype='GPFeatureLayer'
                , parameterType='Required'
                , direction='Input'
            )
        )
        start = arcpy.Parameter(
            displayName='Year'
            , name='year'
            , datatype='GPLong'
            , parameterType='Required'
            , direction='Input'
        )
        #start.filter.list = year_filter
        params.append(start)
        outfile = arcpy.Parameter(
            displayName='Output CSV File'
            , name='outfile'
            , datatype='DEFile'
            , parameterType='Required'
            , direction='Output'
        )
        outfile.filter.list = ['csv']
        params.append(outfile)
        versionparam = arcpy.Parameter(
            displayName='API Version to use'
            , name='version'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Output'
        )
        versionparam.filter.list = ['v2','v3']
        versionparam.value = 'v3'
        params.append(versionparam)
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        params = {p.name: p.valueAsText for p in parameters}
        data = self.run(int(params['year']), params['area'],params['version'])
        with open(params['outfile'], 'wb') as outfile:
            csv_w = csv.DictWriter(outfile, fieldnames=[d.strip() for d in data[0].keys()])
            csv_w.writeheader()
            csv_w.writerows([{k.strip():v.strip() for k,v in d.iteritems()} for d in data])
        return

    @staticmethod
    def run(year, area, api_version='v3'):
        api_url_v3 = 'http://developer.nrel.gov/api/solar/nsrdb_psm3_download.csv'
        api_url_v2 = 'http://developer.nrel.gov/api/solar/nsrdb_0512_download.csv'

        api_attributes_v3 = 'ghi,air_temperature,relative_humidity'
        api_attributes_v2 = 'ghi,surface_air_temperature_nwp,surface_relative_humidity_nwp'

        if api_version=='v3':
            url = api_url_v3
            attributes = api_attributes_v3
        else:
            url = api_url_v2
            attributes = api_attributes_v2

        # todo: add progess bar
        # get centroid lat/lon
        extent = arcpy.Describe(area).extent.projectAs(arcpy.SpatialReference(4326))
        lon, lat = (extent.XMax + extent.XMin) / 2, (extent.YMax + extent.YMin) / 2
        request_dict = {
            'wkt': 'POINT(%s %s)' % (lon, lat)
            , 'names': str(year)
            , 'leap_day': 'false'
            , 'interval': '60'
            , 'utc': 'false'
            , 'full_name': 'Steven Porter'
            , 'email': 'snowballsteve3@gmail.com'
            , 'affiliation': 'Ohio University'
            , 'mailing_list': 'false'
            , 'api_key': 'OC2vFKdZS4vG4QsfUav8C0rvQk29ViPP38H3Vvk3'
            #, 'api_key': '6e0fW7GkuQvNa9JdCM6oiAVwSCXpsSWUSm0qfu66'
            , 'attributes': attributes
            , 'reason': 'ArcGIS Tool'
        }
        request_url = url + '?'
        for k, v in request_dict.iteritems():
            request_url += '&%s=%s' % (k, v)
        #arcpy.AddMessage(request_url)
        retries = 0
        success = False
        from time import sleep
        while retries <= 10 and not success:
            try:
                response = urllib2.urlopen(request_url.replace(' ', '%20'))
                success = True
            except urllib2.HTTPError as e:
                retries += 1
                arcpy.AddWarning(
                    'Failed to retrieve solar year %s with code %s: %s (attemp %d/10)' % (str(year), e.code, http_error[e.code], retries))
                sleep(10)
                if retries > 10:
                    raise DataFetchException("Could not fetch solar data after 10 tries.")


        data = response.read()
        raw_nsrdb = os.path.join(os.path.dirname(os.path.realpath(area)),'raw_nsrdb.txt')
        with open(raw_nsrdb,'w') as f:
            f.write(data)

        data = data.split('\n')
        data = [x + '\n' for x in data]
        data = [x.split(',') for x in data[2:]]
        header = data[0]
        data = [{k: v for k, v in zip(header, x)} for x in data[1:]]
        return data


class DownloadSoilGML(object):
    def __init__(self):
        self.label = "Download: Soil GML"
        self.description = "Downloads gml data from the soil data mart WFS server"
        self.canRunInBackground = False

    def getParameterInfo(self):
        params = []
        params.append(
            arcpy.Parameter(
                displayName='Project Area'
                , name='area'
                , datatype='GPFeatureLayer'
                , parameterType='Required'
                , direction='Input'
            )
        )
        output = arcpy.Parameter(
                displayName='Output GML'
                , name='output_gml'
                , datatype='DEFile'
                , parameterType='Required'
                , direction='Output'
        )
        output.filter.list = ['gml']
        params.append(output)
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        params = {p.name:p.valueAsText for p in parameters}
        self.run(params['area'], params['output_gml'])
        return

    @staticmethod
    def run(project_area, outfile):
        project_describe = arcpy.Describe(project_area)
        project_extent = project_describe.extent.projectAs(arcpy.SpatialReference(4326))
        extent = (project_extent.XMin,project_extent.YMin,project_extent.XMax, project_extent.YMax )
        getSoilPolyGML(extent, outfile)


class ConvertSoilGMLtoShp(object):
    def __init__(self):
        self.label = "Process: Convert Soil GML to SHP"
        self.description = "converts the soil gml data to a shapefile"
        self.canRunInBackground = False

    def getParameterInfo(self):
        params = []
        params.append(
            arcpy.Parameter(
                displayName='Project Area'
                , name='area'
                , datatype='GPFeatureLayer'
                , parameterType='Required'
                , direction='Input'
            )
        )
        gml = arcpy.Parameter(
                displayName='Output GML'
                , name='gml'
                , datatype='DEFile'
                , parameterType='Required'
                , direction='Input'
        )
        gml.filter.list = ['gml']
        params.append(gml)
        params.append(
            arcpy.Parameter(
                displayName='Output Shapefile'
                , name='shp'
                , datatype='DEShapefile'
                , parameterType='Required'
                , direction='Output'
            )
        )
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        params = {p.name:p.valueAsText for p in parameters}
        self.run(params['gml'], params['shp'], params['area'])
        return

    @staticmethod
    def run(gml, shp, area):
        out_path = os.path.abspath(os.path.join(shp, os.pardir))
        temp_shp = os.path.join(out_path, ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(5))+'.shp')
	arcpy.AddMessage(temp_shp)
	os.chdir(os.path.join(os.path.dirname(os.path.abspath(__file__)),'lib','osgeo'))
        os.environ["GDAL_DATA"] = 'gdaldata'
        os.system('ogr2ogr -f "ESRI Shapefile" -s_srs "+proj=latlong +datum=WGS84 +axis=neu +wktext" -t_srs EPSG:4326 %s %s' %(temp_shp, gml))
        time.sleep(20)
	if not arcpy.Exists(temp_shp):
	   raise DataFetchException('OGR2OGR did not produce shapefile conversion')
        arcpy.Clip_analysis(temp_shp, area, shp)
        arcpy.Delete_management(temp_shp)


class GetSoilAWSTable(object):
    def __init__(self):
        self.label = "Download: Soil AWC"
        self.description = ""
        self.canRunInBackground = False

    def getParameterInfo(self):
        params = []
        params.append(
            arcpy.Parameter(
                displayName='Soil Shapefile'
                , name='shp'
                , datatype='DEShapefile'
                , parameterType='Required'
                , direction='Input'
            )
        )
        csv_param = arcpy.Parameter(
                displayName='Output CSV'
                , name='csv'
                , datatype='DEFile'
                , parameterType='Required'
                , direction='Output'
        )
        csv_param.filter.list = ['csv']
        params.append(csv_param)
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        params = {p.name:p.valueAsText for p in parameters}
        self.run({row.getValue('mukey') for row in arcpy.SearchCursor(params['shp'])}, params['csv'])
        return

    @staticmethod
    def run(list_mukey, output_csv):
        data = {
            'query':'select mukey, aws0100wta from muaggatt where mukey IN (%s)' %','.join(["'"+str(x)+"'" for x in list_mukey])
            ,'format':'JSON'
        }
        data = urllib.urlencode(data)
        req = urllib2.Request('https://SDMDataAccess.sc.egov.usda.gov/Tabular/post.rest', data=data)
        try:
            response = urllib2.urlopen(req)

        except urllib2.HTTPError as e:
            arcpy.AddError(
                'Failed to retrieve soil aws with code %s: %s' % (e.code, http_error[e.code]))
            raise

        def cast_aws(x):
            if x:
                return float(x)
            else:
                return None
        result_json = response.read()
        result_dicts = [{'mukey':mukey, 'aws':cast_aws(aws)} for mukey, aws in json.loads(result_json)['Table']]
        csv_w = csv.DictWriter(open(output_csv, 'wb'), fieldnames = ['mukey','aws'])
        csv_w.writeheader()
        csv_w.writerows(result_dicts)
        del csv_w


class CreateSoilRaster(object):
    def __init__(self):
        self.label = "Process: Create Soil Raster"
        self.description = ""
        self.canRunInBackground = False

    def getParameterInfo(self):
        params = []
        params.append(
        arcpy.Parameter(
                displayName='Project Area'
                , name='area'
                , datatype='GPFeatureLayer'
                , parameterType='Required'
                , direction='Input'
            )
        )
        params.append(
        arcpy.Parameter(
                displayName='Project Folder'
                , name='output_dir'
                , datatype='DEFolder'
                , parameterType='Required'
                , direction='Input'
            )
        )
        params.append(
            arcpy.Parameter(
                displayName='DEM'
                , name='dem'
                , datatype='DERasterDataset'
                , parameterType='Required'
                , direction='Input'
            )
        )
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        params = {p.name:p.valueAsText for p in parameters}
        #Step 1, create soils shapefile
        gml_path = os.path.join(params['output_dir'], 'soils.gml')
        arcpyDeleteIfExists(gml_path)
        DownloadSoilGML.run(params['area'], gml_path)

        #step 2, convert to shapefile for area
        shp_path = os.path.join(params['output_dir'], 'soils.shp')
        arcpyDeleteIfExists(shp_path)
        ConvertSoilGMLtoShp.run(gml_path, shp_path, params['area'])

        #step 3, pull aws table
        aws_csv = os.path.join(params['output_dir'], 'soil_awc.csv')
        arcpyDeleteIfExists(aws_csv)
        GetSoilAWSTable.run({row.getValue('mukey') for row in arcpy.SearchCursor(shp_path)}, aws_csv)

        #step 3, join into new shapefile
        soils_lyr = arcpy.MakeFeatureLayer_management(shp_path, 'soils_layer')
        soils_aws_shp = os.path.join(params['output_dir'], 'soils_awc.shp')
        arcpyDeleteIfExists(soils_aws_shp)

        #windows schema hack cause arcgis can't read csv files right

        with open(os.path.join(params['output_dir'],'schema.ini'), 'wb') as schema:
                schema.write("[%s]" %(os.path.basename(aws_csv)) + '\r\n')
                schema.write("Col1=mukey Text")

        arcpy.AddJoin_management(soils_lyr, 'mukey', aws_csv, 'mukey')
        arcpy.Select_analysis(soils_lyr, soils_aws_shp)

        #project soils layer
        soils_projected  = arcpy.Project_management(soils_aws_shp, os.path.join(params['output_dir'], 'soils_awc_projected.shp'), arcpy.Describe(params['area']).spatialReference)
        arcpy.AddField_management(soils_projected, 'value', 'DOUBLE')
        arcpy.CalculateField_management(soils_projected, 'value', '!soil_awc_1!', 'PYTHON')
        #step 4, rasterize
        tif = os.path.join(params['output_dir'], 'soils_awc.tif')
        arcpyDeleteIfExists(tif)
        dem = arcpy.Raster(params['dem'])
        arcpy.PolygonToRaster_conversion(soils_projected, 'value', tif, cell_assignment='MAXIMUM_COMBINED_AREA',
                                         cellsize=min(dem.meanCellHeight,dem.meanCellWidth))
        arcpy.DefineProjection_management(tif, arcpy.Describe(params['area']).spatialReference)

        #cleanup
        arcpy.Delete_management(shp_path)
        arcpy.Delete_management(gml_path)
        os.remove(aws_csv)
        os.remove(os.path.join(params['output_dir'],'schema.ini'))
        arcpy.Delete_management(soils_aws_shp)
        arcpy.Delete_management(soils_projected)



        return


class SolarAnalystPointCSV(object):
    def __init__(self):
        self.label = "Process: Solar Analyst for Centroid Point"
        self.description = "Runs Solar analyst and outputs a nice clean csv"
        self.canRunInBackground = False

    def getParameterInfo(self):
        params = list()
        params.append(
            arcpy.Parameter(
                displayName='Digital Elevation Model'
                , name='dem'
                , datatype='DERasterDataset'
                , parameterType='Required'
                , direction='Input'
            )
        )
        params.append(
            arcpy.Parameter(
                displayName='Project Area'
                , name='area'
                , datatype='GPFeatureLayer'
                , parameterType='Required'
                , direction='Input'
            )
        )
        t = arcpy.Parameter(
            displayName='Transmittivity Proportion'
            , name='t'
            , datatype='GPDouble'
            , parameterType='Required'
            , direction='Input'
        )
        t.filter.range = [0, 1]
        params.append(t)
        d = arcpy.Parameter(
            displayName='Diffuse Proportion'
            , name='d'
            , datatype='GPDouble'
            , parameterType='Required'
            , direction='Input'
        )
        d.filter.range = [0, 1]
        params.append(d)
        outfile = arcpy.Parameter(
            displayName='Outfile'
            , name='outfile'
            , datatype='DEFile'
            , parameterType='Required'
            , direction='Output'
        )
        outfile.filter.list = ['csv']
        params.append(outfile)
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        params = {p.name: p.valueAsText for p in parameters}

        data = self.run(params['dem'], params['area'], params['outfile'], float(params['d']),
                        float(params['t']))

        with open(params['outfile'], 'wb') as outfile:
            csv_w = csv.DictWriter(outfile, fieldnames=data[0].keys())
            csv_w.writeheader()
            csv_w.writerows(data)
        return

    @staticmethod
    def run(dem, area, outfile, d, t):
        arcpy.env.workspace = arcpy.env.scratchWorkspace = os.path.abspath(os.path.join(outfile, os.pardir))
        location_point = arcpy.SearchCursor(area).next().getValue('SHAPE').centroid
        location_point_layer = os.path.join(os.path.abspath(os.path.join(outfile, os.pardir)), 'rad_point.shp')
        if arcpy.Exists(location_point_layer):
            arcpy.Delete_management(location_point_layer)
        location_point_layer = arcpy.CreateFeatureclass_management(
            os.path.abspath(os.path.join(location_point_layer, os.pardir)), 'rad_point.shp', 'POINT')
        in_cur = arcpy.InsertCursor(location_point_layer)
        row = in_cur.newRow()
        row.setValue('SHAPE', location_point)
        in_cur.insertRow(row)
        del in_cur
        temp_rad_file = os.path.join(os.path.abspath(os.path.join(outfile, os.pardir)), 'temp_rad.shp')
        if arcpy.Exists(temp_rad_file):
            arcpy.Delete_management(temp_rad_file)
        arcpy.sa.PointsSolarRadiation(
            dem
            , location_point_layer
            , temp_rad_file
            , sky_size=512
            , time_configuration=arcpy.sa.TimeWholeYear(2015)
            , each_interval=True
            , slope_aspect_input_type='FLAT_SURFACE'
            , diffuse_proportion=d
            , transmittivity=t
            , azimuth_divisions = 16
            , diffuse_model_type='STANDARD_OVERCAST_SKY'
        )
        row = arcpy.SearchCursor(temp_rad_file).next()
        data = []
        for n in range(12):
            data.append({
                'month': n + 1
                , 'arc_ghi': row.getValue('T' + str(n))
                , 'd': d
                , 't': t
            })

        arcpy.Delete_management(temp_rad_file)
        arcpy.Delete_management(location_point_layer)

        return data


class SolarAnalystPointCSV_Multiple(object):
    def __init__(self):
        self.label = "Process: Solar Analyst Point for All Possible D/T Combinations"
        self.description = "Runs solar analyst for point and merges into 1 csv file using all combination of Diffuse and Transmitivity Porportions"
        self.canRunInBackground = False

    def getParameterInfo(self):
        params = list()
        params.append(
            arcpy.Parameter(
                displayName='Digital Elevation Model'
                , name='dem'
                , datatype='DERasterDataset'
                , parameterType='Required'
                , direction='Input'
            )
        )
        params.append(
            arcpy.Parameter(
                displayName='Project Area'
                , name='area'
                , datatype='GPFeatureLayer'
                , parameterType='Required'
                , direction='Input'
            )
        )
        outfile = arcpy.Parameter(
            displayName='Output CSV File'
            , name='outfile'
            , datatype='DEFile'
            , parameterType='Required'
            , direction='Output'
        )
        outfile.filter.list = ['csv']
        params.append(outfile)
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        params = {p.name: p.valueAsText for p in parameters}
        self.run(params['dem'], params['area'], params['outfile'])

    @staticmethod
    def run(dem, project_area, outfile):
        data = []
        for d in [float(x) / 10 for x in range(1, 10)]:
            for t in [float(x) / 10 for x in range(1, 10)]:
                data.extend(SolarAnalystPointCSV.run(dem, project_area, outfile, d, t))

        with open(outfile, 'wb') as outfile:
            csv_w = csv.DictWriter(outfile, fieldnames=data[0].keys())
            csv_w.writeheader()
            csv_w.writerows(data)

        return


class FindBestDT(object):
    def __init__(self):
        self.label = "Process: Find best Diffuse/Transmittivity Numbers"
        self.description = "Using solar analyst results, and nsrdb data, find the best D and T values for each month"
        self.canRunInBackground = False

    def getParameterInfo(self):
        params = list()
        params.append(arcpy.Parameter(
            displayName='Solar Analyst CSV File'
            , name='solar_analyst'
            , datatype='DEFile'
            , parameterType='Required'
            , direction='Input'
        ))
        params.append(arcpy.Parameter(
            displayName='NSRDB PSM CSV File'
            , name='nsrdb'
            , datatype='DEFile'
            , parameterType='Required'
            , direction='Input'
        ))
        outfile = arcpy.Parameter(
            displayName='Output CSV File'
            , name='outfile'
            , datatype='DEFile'
            , parameterType='Required'
            , direction='Output'
        )
        outfile.filter.list = ['csv']
        params.append(outfile)
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        return

    def updateMessages(self, parameters):
        return

    @staticmethod
    def get_nsrdb_month(nsrdb_csv, year, month):
        data = []
        for x in csv.DictReader(open(nsrdb_csv, 'rb')):
            if x['Year'] == str(year) and x['Month'] == str(month):
                data.append(x)

        return data

    @staticmethod
    def get_best_dt(solar_csv, month, reference_value):
        data = []
        for x in csv.DictReader(open(solar_csv, 'rb')):
            if x['month'] == str(month):
                data.append(x)
        diffs = [(x['d'], x['t'], abs(float(x['arc_ghi']) - float(reference_value))) for x in data]
        best_value = min([x[2] for x in diffs])
        d, t, error = [x for x in diffs if x[2] == best_value][0]
        return d, t, error

    def execute(self, parameters, messages):
        params = {p.name: p.valueAsText for p in parameters}
        self.run(params['nsrdb'], params['solar_analyst'], params['outfile'])

        return

    @staticmethod
    def run(nsrdb_csv, solar_analyst_csv, outfile):
        # check that the years match in both datasets
        years_nsrdb = set([int(x['Year']) for x in csv.DictReader(open(nsrdb_csv, 'rb')) if x['Year'] != ''])
        data = []

        for y in years_nsrdb:
            for m in range(1, 13):
                ref_value = sum([float(x['GHI']) for x in FindBestDT.get_nsrdb_month(nsrdb_csv, y, m)])
                best_d, best_t, error = FindBestDT.get_best_dt(solar_analyst_csv, m, ref_value)
                data.append(
                    {
                        'year': y
                        , 'month': m
                        , 'd': best_d
                        , 't': best_t
                        , 'error': error
                    }
                )
        with open(outfile, 'wb') as outfile:
            csv_w = csv.DictWriter(outfile, fieldnames=data[0].keys())
            csv_w.writeheader()
            csv_w.writerows(data)


class SolarAnalystArea(object):
    def __init__(self):
        self.label = "Process: Solar Analyst Area for best DT"
        self.description = ""
        self.canRunInBackground = False

    def getParameterInfo(self):
        params = list()
        params.append(
            arcpy.Parameter(
                displayName='Best DT CSV'
                , name='bestdt'
                , datatype='DEFile'
                , parameterType='Required'
                , direction='Input'
            )
        )
        params.append(
            arcpy.Parameter(
                displayName='Digital Elevation Model'
                , name='dem'
                , datatype='DERasterDataset'
                , parameterType='Required'
                , direction='Input'
            )
        )
        params.append(
            arcpy.Parameter(
                displayName='Output Folder'
                , name='output_dir'
                , datatype='DEFolder'
                , parameterType='Required'
                , direction='Input'
            )
        )
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        params = {p.name: p.valueAsText for p in parameters}
        self.run(params['dem'], params['bestdt'], params['output_dir'])
        return

    @staticmethod
    def run(dem, best_dt_file, outfolder):
        """
        Using the reference best dt csv file. Tool produces a solar analyst output for each year, for each dt combination needed
        :param dem:
        :param best_dt_file:
        :param outfolder:
        :return:
        """
        check_folder(os.path.join(outfolder, 'solar'))
        arcpy.env.workspace = arcpy.env.scratchWorkspace = os.path.join(outfolder, 'solar')
        csv_r = csv.DictReader(open(best_dt_file, 'rb'))
        data = set([(row['d'], row['t']) for row in csv_r])
        message = 'Found %d D/T combinations to run' % (len(data))
        arcpy.SetProgressor('step', message, 0, len(data))
        arcpy.AddMessage(message)

        for i, datum in enumerate(data):
            d, t = datum
            d = float(d)
            t = float(t)
            arcpy.SetProgressorLabel('Running %d/%d' % (i, len(data)))
            out_path = 'solar_d%d_t%d.tif' % (d * 10, t * 10)
            if arcpy.Exists(out_path):
                arcpy.AddMessage('%s already exists, skipping' %out_path)
            else:
                solar_output = arcpy.sa.AreaSolarRadiation(dem, azimuth_divisions=16,
                                                           time_configuration=arcpy.sa.TimeWholeYear(2015),
                                                           diffuse_proportion=d,
                                                           transmittivity=t, each_interval='INTERVAL',
							   diffuse_model_type='STANDARD_OVERCAST_SKY')
                solar_output.save(out_path)
            arcpy.SetProgressorPosition()


class DownloadNED(object):
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Download: Elevation Model"
        self.description = ""
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Define parameter definitions"""
        params = []
        params.append(
        arcpy.Parameter(
                displayName='Project Area'
                , name='area'
                , datatype='GPFeatureLayer'
                , parameterType='Required'
                , direction='Input'
            )
        )
        params.append(
        arcpy.Parameter(
                displayName='Output Folder'
                , name='output_dir'
                , datatype='DEFolder'
                , parameterType='Required'
                , direction='Input'
            )
        )
        return params

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
        return True

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""
        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""
        return

    def execute(self, parameters, messages):
        """The source code of the tool."""
        params = {p.name:p.valueAsText for p in parameters}
        area = params['area']
        outdir = params['output_dir']
        #get extent
        extent = arcpy.Describe(area).extent.projectAs(arcpy.SpatialReference(4326))
        lats = range(int(extent.YMin)+1, int(extent.YMax)+2)
        longs = range(int(extent.XMin)-1, int(extent.XMax))

        grids = [zip(x,longs) for x in itertools.permutations(lats,len(longs))]
        grids = [item for sublist in grids for item in sublist]

        fetched_files = []
        rasters = []
        trash_files = []
        for lat, lng in grids:
            retries = 0
            success = False
            while retries < 9 and not success:
                #'https://prd-tnm.s3.amazonaws.com/StagedProducts/Elevation/13/IMG/n41w083.zip'
                #'https://prd-tnm.s3.amazonaws.com/StagedProducts/Elevation/13/IMG/USGS_NED_13_n43w074_IMG.zip'
                url_get1 = 'https://prd-tnm.s3.amazonaws.com/StagedProducts/Elevation/13/IMG/n%02dw%03d.zip' %(lat,abs(lng))
                url_get2 =  'https://prd-tnm.s3.amazonaws.com/StagedProducts/Elevation/13/IMG/USGS_NED_13_n%02dw%03d_IMG.zip' %(lat,abs(lng))

                #try url 1 if odd, url2 if even
                if retries%2 == 0:
                    url_get = url_get2
                    zip_path = 'USGS_NED_13_n%02dw%03d_IMG.zip' % (lat, abs(lng))
                else:
                    url_get = url_get1
                    zip_path = 'n%02dw%03d.zip' % (lat, abs(lng))
                arcpy.AddMessage(url_get)
                out_path = os.path.join(outdir, zip_path)
                for suffix in ['_13_meta.shp','_13_meta.txt','_13_meta.xml','_13_meta.html']:
                    trash_files.append(os.path.join(outdir, 'n%02dw%03d%s' %(lat,abs(lng), suffix)))
                trash_files.append(os.path.join(outdir, 'imgn%02dw%03d_13_thumb.jpg' %(lat,abs(lng))))
                try:
                    urllib.urlretrieve(url_get, out_path)

                except urllib.error.HTTPError as e:
                    arcpy.AddWarning(
                        'Failed to retrieve %s with code %s: %s' % (url_get,e.code, http_error[e.code]))
                    retries +=1

                #extract zip file
                try:
                    zip_ref = zipfile.ZipFile(out_path, 'r')
                    zip_ref.extractall(outdir)
                    zip_ref.close()
                    success=True
                    fetched_files.append(out_path)
                    if url_get == url_get1:
                        rasters.append('imgn%02dw%03d_13.img' % (lat, abs(lng)))
                    else:
                        rasters.append('USGS_NED_13_n%02dw%03d_IMG.img' % (lat, abs(lng)))
                except zipfile.BadZipfile:
                    arcpy.AddWarning('Failed to open zip file, bad download, trying again')
                    retries +=1






        arcpy.MosaicToNewRaster_management([os.path.join(outdir, path) for path in rasters] ,outdir, 'raw_dem.tif', number_of_bands=1, pixel_type='32_BIT_FLOAT')

        #clean up in between data
        def a_del(f):
            arcpy.Delete_management(f)

        map(a_del, fetched_files)
        map(a_del, [os.path.join(outdir,r) for r in rasters])
        map(a_del, trash_files)
        a_del(os.path.join(outdir,'ned_13arcsec_g.shp'))
        os.remove(os.path.join(outdir, 'readme.pdf'))
        os.remove(os.path.join(outdir, 'NED_DataDictionary.url'))

        return
