import arcpy
import os
import csv
import shutil
from lib.geo_methods import *


class MiscException(Exception):
    """
    Custom exception that can tell you the tool that caused it as well as include the error message
    """

    def __init__(self, tool, message):
        new_message = tool.label + ' error: ' + message
        arcpy.AddError(new_message)
        super(MiscException, self).__init__(new_message)

class Toolbox(object):
    def __init__(self):
        """Define the toolbox (the name of the toolbox is the name of the
        .pyt file)."""
        self.label = "Water Balance Tool: Misc"
        self.alias = "Misc"

        # List of tool classes associated with this toolbox
        self.tools = [SnapRaster, ExtractAndProjectRaster, LoadData, DataCleanup,LoadDataFromValues, CheckDivide]


class SnapRaster(object):
    def __init__(self):
        self.label = "Snap Raster"
        self.description = "Snaps the input raster to the reference raster after resampling cellsize"
        self.canRunInBackground = False

    def getParameterInfo(self):
        params = list()
        params.append(
            arcpy.Parameter(
                displayName='Input Raster'
                , name='input_raster'
                , datatype='DERasterDataset'
                , parameterType='Required'
                , direction='Input'
            )
        )
        params.append(
            arcpy.Parameter(
                displayName='Reference Raster'
                , name='reference_raster'
                , datatype='DERasterDataset'
                , parameterType='Required'
                , direction='Input'
            )
        )
        params.append(
            arcpy.Parameter(
                displayName='Output Raster'
                , name='output_raster'
                , datatype='DERasterDataset'
                , parameterType='Required'
                , direction='Output'
            )
        )
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        params = {p.name: p.valueAsText for p in parameters}
        input_raster = arcpy.Raster(params['input_raster'])
        reference_raster = arcpy.Raster(params['reference_raster'])
        output_raster = params['output_raster']
        return self.run(input_raster, reference_raster, output_raster)

    @staticmethod
    def run(input_raster, reference_raster, output_raster, skip_resamp=False):
        """
        Snaps an input raster to the reference raster. Expects that the grid size is already the same and the projection must match
        :param input_raster: raster to be adjusted
        :param reference_raster: raster whose grid is used as a reference
        :param output_raster: location to save new raster
        :return: output_raster
        """
        # arcpy.env.workspace = os.path.abspath(os.path.join(output_raster, os.pardir))
        prev_workspace = arcpy.env.workspace
        arcpy.env.workspace = os.path.dirname(os.path.realpath(output_raster))

        # check that rasters are already arcpy.raster, if not make it so.
        if not arcpy.Raster == type(reference_raster):
            raise MiscException(SnapRaster(), 'Reference Raster is not of type arcpy.raster')
        if not arcpy.Raster == type(input_raster):
            raise MiscException(SnapRaster(), 'Input Raster is not of type arcpy.raster')

        # check to ensure same projection
        if not input_raster.spatialReference.Name == reference_raster.spatialReference.Name:
            arcpy.AddError('Input Rasters must be in the same projection (%s,%s)' %(input_raster.spatialReference.Name,reference_raster.spatialReference.Name ))
            raise MiscException(SnapRaster, 'Error: Input Rasters must be in the same projection')

        # resample input to reference cell size

        cell_size = (reference_raster.meanCellWidth, reference_raster.meanCellHeight)
        if not abs(input_raster.meanCellWidth - cell_size[0]) < 0.00001 and not abs(
                        input_raster.meanCellHeight - cell_size[1]) < 0.00001 and not skip_resamp:
            resamp_temp = 'resamp.tif'
            resamp_temp = arcpy.Resample_management(input_raster, resamp_temp, '%f %f' % cell_size, 'NEAREST')
            resampled_raster = arcpy.Raster(resamp_temp)
        else:
            resampled_raster = input_raster

        # shift resampled raster
        dx1 = resampled_raster.extent.XMin % resampled_raster.meanCellWidth
        dy1 = resampled_raster.extent.YMin % resampled_raster.meanCellHeight
        dx2 = reference_raster.extent.XMin % reference_raster.meanCellWidth
        dy2 = reference_raster.extent.YMin % reference_raster.meanCellHeight
        out_raster_shift = arcpy.Shift_management(resampled_raster, 'shift.tif', dx2 - dx1, dy2 - dy1,
                                            in_snap_raster=reference_raster)
        out_raster = arcpy.CopyRaster_management(out_raster_shift, output_raster)
        arcpy.env.workspace = prev_workspace
        arcpy.Delete_management(out_raster_shift)

        return out_raster


class ExtractAndProjectRaster(object):
    def __init__(self):
        self.label = "Extract and Project Raster by Mask"
        self.description = "Extracts input raster based on a mask, or an extent"
        self.canRunInBackground = False

    def getParameterInfo(self):
        params = list()
        params.append(
            arcpy.Parameter(
                displayName='Input Raster'
                , name='input_raster'
                , datatype='DERasterDataset'
                , parameterType='Required'
                , direction='Input'
            )
        )
        params.append(
            arcpy.Parameter(
                displayName='Mask'
                , name='mask'
                , datatype='GPFeatureLayer'
                , parameterType='Required'
                , direction='Input'
            )
        )
        extent_only = arcpy.Parameter(
            displayName='Use Extent Only'
            , name='extent_only'
            , datatype='GPBoolean'
            , parameterType='Optional'
            , direction='Input'
        )
        params.append(
            arcpy.Parameter(
                displayName='Out Raster'
                , name='out_raster'
                , datatype='DERasterDataset'
                , parameterType='Required'
                , direction='Output'
            )
        )
        extent_only.value = True
        params.append(extent_only)
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        params = {p.name: p.valueAsText for p in parameters}

        return self.run(params['input_raster'], params['mask'], params['out_raster'], params['extent_only'] == 'true')

    @staticmethod
    def run(input_raster, mask, out_raster, extent_only=True):
        arcpy.env.workspace = os.path.abspath(os.path.join(out_raster, os.pardir))

        input_raster = arcpy.Raster(input_raster)

        # extract oversized raster
        mask_describe = arcpy.Describe(mask)
        mask_extent = mask_describe.extent
        mask_extent_buffer = buffer_extent_by_perc(mask_extent, 0.1)
        envelope = extent_to_envelope(mask_extent_buffer.projectAs(input_raster.spatialReference))
        extracted_buffer = arcpy.Clip_management(input_raster, envelope, 'temp_ex.tif', maintain_clipping_extent=False)

        # project to masks projection
        projected_raster = arcpy.Raster(
            arcpy.ProjectRaster_management(extracted_buffer, 'temp_proj.tif', mask_describe.spatialReference,
                                           'BILINEAR'))

        # extract final raster
        if extent_only:
            out_raster = arcpy.Clip_management(projected_raster, extent_to_envelope(mask_extent), out_raster,
                                               maintain_clipping_extent=False)
        else:
            out_raster = arcpy.Clip_management(projected_raster, mask, out_raster, maintain_clipping_extent=False,
                                               clipping_geometry=True)

        arcpy.Delete_management(extracted_buffer)
        arcpy.Delete_management(projected_raster)
        return out_raster


class LoadData(object):
    def __init__(self):
        self.label = "Load Custom Raster Data"
        self.description = ""
        self.canRunInBackground = False

    def getParameterInfo(self):
        params = []
        params.append(
            arcpy.Parameter(
                displayName='Project Folder'
                , name='project_folder'
                , datatype='DEFolder'
                , parameterType='Required'
                , direction='Input'
            )
        )
        params.append(
            arcpy.Parameter(
                displayName='Project area'
                , name='project_area'
                , datatype='DEShapefile'
                , parameterType='Required'
                , direction='Input'
            )
        )
        params.append(
            arcpy.Parameter(
                displayName='Project DEM'
                , name='dem'
                , datatype='DERasterDataset'
                , parameterType='Required'
                , direction='Input'
            )
        )
        params.append(arcpy.Parameter(
            displayName='Load Data CSV File'
            , name='load'
            , datatype='DEFile'
            , parameterType='Required'
            , direction='Input'
        ))

        prefix = arcpy.Parameter(
            displayName='Prefix to process'
            , name='prefix'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )
        prefix.filter.list = ['All']
        params.append(prefix)
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        params = {p.name:p for p in parameters}
        if params['load'].altered:
            params['prefix'].filter.list = list({row['prefix'] for row in csv.DictReader(open(params['load'].valueAsText, 'r'))})+['All']
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        params = {p.name:p.valueAsText for p in parameters}
        self.run(params['load'], params['project_folder'], params['project_area'], params['dem'], params['prefix'])
        return

    def run(self, load_csv, project_folder, project_area, dem, prefix_to_process='All'):
        arcpy.workspace = project_folder
        arcpy.env.overwriteOutput = True
        grid = arcpy.Raster(dem)
        if not arcpy.Exists(grid):
            raise MiscException(self,'Cannot find reference DEM in project folder')

        for row in csv.DictReader(open(load_csv, 'r')):
            # extract and snap
            if row['prefix'] == prefix_to_process or prefix_to_process == 'All':
                path, prefix, y, m = (row['path'], row['prefix'], int(row['year']), int(row['month']))
                arcpy.AddMessage('Processing %s for %d-%02d' %(prefix, y, m))
                if not arcpy.Exists(path):
                    arcpy.AddWarning('File Not Found, skipping: %s' %path)
                else:
                    arcpy.env.cellSize = 'MINOF'
                    rast = arcpy.Raster(path)
                    temp2 = grid * 0 + rast
                    temp2_path = os.path.join(project_folder,'temp2.tif')
                    temp2.save(temp2_path)
                    del temp2
                    arcpy.env.cellSize = 'MAXOF'
                    ExtractAndProjectRaster.run(temp2_path,project_area,os.path.join(project_folder,'temp.tif'))
                    check_folder(os.path.join(project_folder, prefix))
                    SnapRaster.run(arcpy.Raster(os.path.join(project_folder,'temp.tif')),grid,os.path.join(project_folder, prefix,'%s_%d_%02d.tif' %(prefix, y, m)))
                    arcpy.Delete_management(temp2_path)
                    if arcpy.Exists(os.path.join(project_folder,'temp.tif')):
                        arcpy.Delete_management(os.path.join(project_folder,'temp.tif'))

class LoadDataFromValues(object):
    def __init__(self):
        self.label = "Load Custom Value Data"
        self.description = ""
        self.canRunInBackground = False

    def getParameterInfo(self):
        params = []
        params.append(
            arcpy.Parameter(
                displayName='Project Folder'
                , name='project_folder'
                , datatype='DEFolder'
                , parameterType='Required'
                , direction='Input'
            )
        )
        params.append(
            arcpy.Parameter(
                displayName='Project DEM'
                , name='dem'
                , datatype='DERasterDataset'
                , parameterType='Required'
                , direction='Input'
            )
        )
        params.append(arcpy.Parameter(
            displayName='Load Data CSV File'
            , name='load'
            , datatype='DEFile'
            , parameterType='Required'
            , direction='Input'
        ))

        prefix = arcpy.Parameter(
            displayName='Prefix to process'
            , name='prefix'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )
        prefix.filter.list = ['All']
        params.append(prefix)
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        params = {p.name:p for p in parameters}
        if params['load'].altered:
            params['prefix'].filter.list = list({row['prefix'] for row in csv.DictReader(open(params['load'].valueAsText, 'r'))})+['All']

        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        params = {p.name:p.valueAsText for p in parameters}
        self.run(params['load'], params['project_folder'], params['dem'], params['prefix'])
        return

    def run(self, load_csv, project_folder, grid_path, prefix_to_process='All'):
        arcpy.workspace = project_folder
        arcpy.env.overwriteOutput = True
        if not arcpy.Exists(grid_path):
            raise MiscException(self,'Cannot find reference grid in project folder')

        grid = arcpy.Raster(grid_path)
        for row in csv.DictReader(open(load_csv, 'r')):
            # extract and snap
            if row['prefix'] == prefix_to_process or prefix_to_process == 'All':
                value, prefix, y, m = (row['value'], row['prefix'], int(row['year']), int(row['month']))
                arcpy.AddMessage('Processing %s for %d-%02d' %(prefix, y, m))
                if not os.path.exists(os.path.join(project_folder, prefix)):
                    os.makedirs(os.path.join(project_folder, prefix))
                result = (grid - grid) + 1 * float(value)
                result.save(os.path.join(project_folder, prefix,'%s_%d_%02d.tif' %(prefix, y, m)))




class DataCleanup(object):
    def __init__(self):
        self.label = "Data Cleanup"
        self.description = ""
        self.canRunInBackground = False

    def getParameterInfo(self):
        params = []
        params.append(
            arcpy.Parameter(
                displayName='Project Folder'
                , name='project_folder'
                , datatype='DEFolder'
                , parameterType='Required'
                , direction='Input'
            )
        )
        params.append(
            arcpy.Parameter(
                displayName='Datasets to delete'
                , name='subfolders'
                , datatype='GPString'
                , parameterType='Required'
                , direction='Input'
                ,multiValue=True
            )
        )
        out_csv = arcpy.Parameter(
                displayName='Save values to CSV'
                , name='out_csv'
                , datatype='DEFile'
                , parameterType='Optional'
                , direction='Output'
            )
        params.append(out_csv)
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        params = {p.name:p for p in parameters}
        folder =params['project_folder']
        subfolder=params['subfolders']
        if folder.altered:
            subfolder.filter.list = [o for o in os.listdir(folder.valueAsText) if os.path.isdir(os.path.join(folder.valueAsText,o)) and o not in ['info','solar']]
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        params = {p.name:p.valueAsText for p in parameters}
        project_folder = params['project_folder']
        out_csv = params['out_csv']

        for subfolder in params['subfolders'].split(';'):
            if subfolder == 'solar':
                shutil.rmtree(os.path.join(project_folder, subfolder))
            else:
                arcpy.AddMessage(subfolder)
                for f in os.listdir(os.path.join(project_folder,subfolder)):
                    if f.endswith('.tif'):
                        split_out = f.split('_')
                        if len(split_out) == 3:
                            prefix, year, month = split_out
                            month = month.split('.')[0]
                            year = year
                            row = {
                                'prefix':prefix
                                ,'year':year
                                ,'month':month
                                ,'value':arcpy.Raster(os.path.join(project_folder, subfolder, f)).mean
                            }
                            if not os.path.exists(out_csv):
                                csv_w = csv.DictWriter(open(out_csv, 'ab'), fieldnames = row.keys())
                                csv_w.writeheader()
                                csv_w.writerow(row)
                                del csv_w
                            else:
                                csv_w = csv.DictWriter(open(out_csv, 'ab'), fieldnames=row.keys())
                                csv_w.writerow(row)
                                del csv_w

                        arcpy.Delete_management(os.path.join(project_folder, subfolder, f))
            shutil.rmtree(os.path.join(project_folder, subfolder))
        return

class CheckDivide(object):
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Check for Divison Errors"
        self.description = ""
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Define parameter definitions"""
        params = []
        params.append(
            arcpy.Parameter(
                displayName='Temp Workspace'
                , name='workspace'
                , datatype='DEWorkspace'
                , parameterType='Required'
                , direction='Input'
            )
        )
        return params

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
        return True

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""
        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""
        return

    def execute(self, parameters, messages):
        """The source code of the tool."""
        params = {p.name: p.valueAsText for p in parameters}
        arcpy.env.workspace = arcpy.env.scratchWorkspace = params['workspace']
        a = 123456.789
        b = 1.23456
        db = 2.345
        while b < a-db:
            a_ras = arcpy.sa.CreateConstantRaster(a, "FLOAT", 1)
            b_ras = arcpy.sa.CreateConstantRaster(b, "FLOAT", 1)
            c_ras = a_ras / b_ras
            c_ras_value = float(arcpy.GetCellValue_management(c_ras, "125 125", "1").getOutput(0))
            c = a / b
            arcpy.AddMessage("%f / %f; error = %f" %(a,b,abs(c - c_ras_value)))
            b = b * db
        return