import arcpy
import os
import csv
import random
import string
from lib.geo_methods import *
from calendar import monthrange
from copy import deepcopy

class WaterBudgetException(Exception):
    """
    Custom exception that can tell you the tool that caused it as well as include the error message
    """

    def __init__(self, tool, message):
        new_message = tool.label + ' error: ' + message
        arcpy.AddError(new_message)
        super(WaterBudgetException, self).__init__(new_message)


class StorageMonth:
    """
    Object data model to handle monthly wbt rasters (or create sub rasters)
    """

    def __init__(self, project_folder, year, month):
        """

        :param project_folder: path where all raster files will be found
        :param year: int
        :param month: int
        """
        self.folder = project_folder
        self.year = int(year)
        self.month = int(month)
        self.pet = arcpy.Raster(os.path.join(self.folder, 'pet\pet_%d_%02d.tif' % (self.year, self.month)))
        self.pet_day = self.pet / monthrange(self.year, self.month)[1]
        self.p = arcpy.Raster(os.path.join(self.folder, 'ppt\ppt_%d_%02d.tif' % (self.year, self.month)))
        self.p_day = self.p / monthrange(self.year, self.month)[1]
        progress_message('Calculating P-PE for %d-%d' % (year, month))
        ppe_path = os.path.join(project_folder, 'ppe\ppe_%d_%02d.tif' % (year, month))
        if arcpy.Exists(ppe_path):
            progress_message('P-PE already calculated, skipping')
            self.ppe = arcpy.Raster(ppe_path)
        else:
            self.ppe = self.p - self.pet
            progress_message('Saving P-PE for %d-%d' % (year, month))
            check_folder(os.path.join(project_folder, 'ppe'))
            self.ppe.save(ppe_path)
        self.ppe_day = self.ppe / monthrange(self.year, self.month)[1]


class Toolbox(object):
    def __init__(self):
        """Define the toolbox (the name of the toolbox is the name of the
        .pyt file)."""
        self.label = "Water Balance Tool: Water Storage Model"
        self.alias = "WaterBudget"

        # List of tool classes associated with this toolbox
        self.tools = [PET, StorageRoutine, StorageDerived, Diagnostics, Checks]


class PET(object):
    def __init__(self):
        self.label = "Calculate PET"
        self.description = "Calculates PET for each month of each year and produces a raster for each."
        self.canRunInBackground = False

    def getParameterInfo(self):
        params = list()
        params.append(
            arcpy.Parameter(
                displayName='Project Folder'
                , name='output_dir'
                , datatype='DEFolder'
                , parameterType='Required'
                , direction='Input'
            )
        )
        params.append(
            arcpy.Parameter(
                displayName='Best DT CSV'
                , name='bestdt'
                , datatype='DEFile'
                , parameterType='Required'
                , direction='Input'
            )
        )
        humid = arcpy.Parameter(
                displayName='Force Humid Calculations'
                , name='humid'
                , datatype='GPBoolean'
                , parameterType='Required'
                , direction='Input'
        )
        humid.value=False
        params.append(humid)
        params.append(
            arcpy.Parameter(
                displayName='NSRDB CSV'
                , name='nsrdb'
                , datatype='DEFile'
                , parameterType='Optional'
                , direction='Input'
            )
        )
        start = arcpy.Parameter(
            displayName='Start Year'
            , name='start_year'
            , datatype='GPLong'
            , parameterType='Required'
            , direction='Input'
        )
        #start.filter.list = year_filter
        params.append(start)
        end = arcpy.Parameter(
            displayName='End Year'
            , name='end_year'
            , datatype='GPLong'
            , parameterType='Required'
            , direction='Input'
        )
        #end.filter.list = year_filter
        params.append(end)
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        params = {p.name: p for p in parameters}
        if params['humid'].altered:
            humid = params['humid'].valueAsText == 'true'
            params['nsrdb'].enabled = not humid
        if params['output_dir'].altered:
            bestdt_path = os.path.join(params['output_dir'].valueAsText,'best_dt.csv')
            if os.path.exists(bestdt_path):
                params['bestdt'].value = bestdt_path
                years = [int(r['year']) for r in csv.DictReader(open(bestdt_path,'r'))]
                params['start_year'].value = min(years)
                params['end_year'].value = max(years)
            else:
                params['bestdt'].value = None
            nsrdb_path = os.path.join(params['output_dir'].valueAsText,'nsrdb.csv')
            if os.path.exists(nsrdb_path):
                params['nsrdb'].value = nsrdb_path
            else:
                params['nsrdb'].value = None
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        params = {p.name: p.valueAsText for p in parameters}
        self.run(params['output_dir'], params['bestdt'], params['nsrdb'], int(params['start_year']), int(params['end_year']), params['humid']=='true')
        return

    def get_dt(self, bestdtcsv, month, year):
        bestdt_data = [d for d in csv.DictReader(open(bestdtcsv, 'rb'))]
        result = [(float(row['d']), float(row['t'])) for row in bestdt_data if
                  row['month'] == str(month) and row['year'] == str(year)]
        if not result:
            raise WaterBudgetException(self, 'Could not find best dt for %s-%s' % (month, year))
        return result[0]

    def run(self, project_folder, bestdtcsv, nsrdb, start_year, end_year, humid=False):
        arcpy.AddMessage('Starting PET')
        arcpy.env.overwriteOutput = True
        check_folder(os.path.join(project_folder,'pet'))
        arcpy.env.workspace = arcpy.env.scratchWorkspace = os.path.join(project_folder,'pet')


        for year in range(start_year, end_year + 1):
            arcpy.AddMessage('Starting %s' %(str(year)))
            #read rel humidity data
            if not humid:
                arcpy.AddMessage('Loading Humidity')
                rel_hum = csv.DictReader(open(nsrdb,'r'))
                rel_hum = [(r['Month'], r['Relative Humidity']) for r in rel_hum if r['Year']==str(year)]
                hum_m = []
                hum_rh = []
                for m,rh in rel_hum:
                    if(m>'' and rh>''):
                        m = int(m)
                        rh = float(rh)
                        hum_m.append(m)
                        hum_rh.append(rh)

                result = {}
                for m in range(1,13):
                    arcpy.AddMessage('Starting %s-%s' %(str(m),str(year)))
                    n = len([x for x in hum_m if x == m])
                    rh = sum([x[1] for x in zip(hum_m, hum_rh) if x[0] == m])
                    result[m] = rh / n
                hum = result
            pets = []
            for month in range(1, 13):
                progress_message('Calculating PET for %d-%02d' % (year, month))
                d, t = self.get_dt(bestdtcsv, month, year)
                arcpy.AddMessage('BestDT %s-%s, loading rasters' % (str(d), str(t)))
                solar = arcpy.Raster(
                    os.path.join(project_folder, 'solar\solar_d%d_t%d.tif\Band_%d' % (d * 10, t * 10, month)))
                temp = arcpy.Raster(os.path.join(project_folder, 'tmean\\tmean_%d_%02d.tif' % (year, month)))
                if temp.mean <= 0:
                    arcpy.AddMessage('Temp < 0')
                    pet = temp * 0
                    #pet = arcpy.sa.CreateConstantRaster(0, "FLOAT", cell_size=temp.meanCellWidth, extent=temp.extent)
                    #temp_proj = arcpy.Describe(temp).spatialReference
                    #arcpy.DefineProjection_management(pet, temp_proj)

                else:
                    arcpy.AddMessage('Calculating PET')
                    if humid:
                        pet = (solar * 0.08598 + 50) * temp / (temp + 15) * 0.013
                    elif hum[month] >= 50:
                        arcpy.AddMessage('%s - %s is a humid month' %(year, month))
                        pet = (solar * 0.08598 + 50) * temp / (temp + 15) * 0.013
                    else:
                        arcpy.AddMessage('%s - %s is a dry month' %(year, month))
                        pet = (solar * 0.08598 + 50) * temp / (temp + 15) * 0.013 * (1+((50 - hum[month]) / 70))

                arcpy.AddMessage('Saving PET')
                pet.save(os.path.join(project_folder,'pet\pet_%d_%02d.tif' % (year, month)))
                pets.append(pet)

            progress_message('Calculating Annual PET for %d' % (year))
            pet_annual = pets[0] + pets[1] + pets[2] + pets[3] + pets[4] + pets[5] + pets[6] + pets[7] + pets[8] + pets[
                9] + pets[10] + pets[11]
            pet_annual.save(os.path.join(project_folder,'pet','pet_%d_annual.tif' % year))
        pass


class StorageRoutine(object):
    def __init__(self):
        self.label = "Run Storage Routine"
        self.description = ""
        self.canRunInBackground = False

    def getParameterInfo(self):
        params = list()
        params.append(
            arcpy.Parameter(
                displayName='Project Folder'
                , name='output_dir'
                , datatype='DEFolder'
                , parameterType='Required'
                , direction='Input'
            )
        )
        start = arcpy.Parameter(
            displayName='Start Year'
            , name='start_year'
            , datatype='GPLong'
            , parameterType='Required'
            , direction='Input'
        )
        #start.filter.list = year_filter
        params.append(start)
        end = arcpy.Parameter(
            displayName='End Year'
            , name='end_year'
            , datatype='GPLong'
            , parameterType='Required'
            , direction='Input'
        )
        #end.filter.list = year_filter
        params.append(end)
        params.append(
            arcpy.Parameter(
                displayName='Debug Points'
                , name='debug_points'
                , datatype='DEShapefile'
                , parameterType='Optional'
                , direction='Input'
            )
        )
        out_csv = arcpy.Parameter(
            displayName='Debug Output'
            , name='debug_csv'
            , datatype='DEFile'
            , parameterType='Optional'
            , direction='Output'
        )
        params.append(out_csv)
        skip = arcpy.Parameter(
                displayName='Year 1 January Storage - Wrapping Option'
                , name='skip_wrap'
                , datatype='GPBoolean'
                , parameterType='Optional'
                , direction='Input'
            )
        skip.value = False
        params.append(skip)
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        return

    def updateMessages(self, parameters):
        params = {p.name: p for p in parameters}
        if (params['start_year'].altered or params['output_dir'].altered) and params['output_dir'].valueAsText > '' and \
                params['start_year'].value:
            possible_previous_storage = os.path.join(params['output_dir'].valueAsText,
                                                     'storage\storage_%s_%s.tif' % (
                                                     int(params['start_year'].valueAsText) - 1, 12))
            if arcpy.Exists(possible_previous_storage):
                params['start_year'].setWarningMessage(
                    "Storage for previous year %s exists, routine use it as starting point" % (
                    int(params['start_year'].valueAsText) - 1))
                if params['skip_wrap'].valueAsText == 'true':
                    params['skip_wrap'].setWarningMessage(
                        "There is an existing storage file for December of %s, wrapping condition will not be met. If you wish to wrap please move or delete this file." % (
                        int(params['start_year'].valueAsText) - 1))
            else:
                params['start_year'].clearMessage()
                params['skip_wrap'].clearMessage()

        return

    def execute(self, parameters, messages):
        arcpy.env.overwriteOutput = True
        params = {p.name: p.valueAsText for p in parameters}
        start = datetime.date(int(params['start_year']), 1, 1)
        end = datetime.date(int(params['end_year']), 12, 31)
        if params['debug_points']:
            self.run(params['output_dir'], start, end, params['debug_points'], params['debug_csv'], skip_wrap=params['skip_wrap']<>'true')
        else:
            self.run(params['output_dir'], start, end, skip_wrap=params['skip_wrap']<>'true')
        return

    def run(self, project_folder, start_date, end_date, debug_points=None, debug_csv=None, skip_wrap=True):
        """
        Runs daily storage calculations from the first day of start_year ot the last day of end_year
        :param project_folder: folder where all rasters are found
        :param start_date: datetime
        :param end_date: datetime
        :return:
        """

        # create empty variables for month objects to use
        current_month = None
        month_full = None
        wrap=False

        # start storage with awc
        arcpy.env.workspace = arcpy.env.scratchWorkspace = project_folder
        arcpy.env.NoData = "PROMOTION"
        check_folder(os.path.join(project_folder,'storage'))

        storage_path = os.path.join(project_folder, 'storage\storage.tif')
        awc = arcpy.Raster(os.path.join(project_folder, 'soils.tif'))

        #set nodata values in awc to 0
        #awc = arcpy.sa.Con(arcpy.sa.IsNull(awc),0,awc,"Value = 1")
        possible_previous_storage = os.path.join(project_folder, 'storage\storage_%s_%s.tif' % (start_date.year - 1, 12))
        if arcpy.Exists(possible_previous_storage):
            storage = arcpy.Raster(possible_previous_storage)
        else:
            storage = awc
            if skip_wrap:
                storage.save(possible_previous_storage)
        storage.save(storage_path)
        skip_month = False

        # iterate for each day, call generator
        dt = start_date
        while dt >= start_date and dt<= end_date:
            #for dt in days(start_date, end_date):
            # start storage with awc
            arcpy.env.workspace = arcpy.env.scratchWorkspace = project_folder
            progress_message('Running storage for %s' % dt)
            if debug_points: storage_start = storage
            # is it the start of month
            if dt.day == 1:
                # copy object to previous and create new object for current month
                if arcpy.Exists(os.path.join(project_folder, 'storage\storage_%d_%02d.tif')):
                    arcpy.Delete_management(os.path.join(project_folder, 'storage\storage_%d_%02d.tif'))

                current_month = StorageMonth(project_folder, dt.year, dt.month)

                if 'month_full' in locals() or 'month_full' in globals():
                    del month_full
                month_full_path = os.path.join(project_folder,'month_full.tif')
                month_full = arcpy.sa.Abs(storage - awc) <= 0.001
                month_full.save(month_full_path)
                month_full_flag = month_full.minimum == month_full.maximum == 1


                del month_full
                arcpy.Delete_management(month_full_path)

                skip_month = month_full_flag and current_month.ppe.minimum >= 0

            if not skip_month:
                # check if storage is full (in which case all cells will be equal to true after boolean test
                storage_full = arcpy.sa.Abs(storage - awc) <= 0.001
                storage_full.save(os.path.join(project_folder,'storage_full.tif'))
                storage_full = storage_full.minimum == storage_full.maximum == 1

                # if ppe has any negative values or storage is not full we need to calculate this day, else it just remains full
                if current_month.ppe.minimum < 0 or not storage_full:
                    ppe = current_month.ppe_day
                    ppe_available = 0.0 + ppe * (storage / awc) * (ppe < 0.0) + ppe * (ppe >= 0.0)
                    storage = 0.0 + (storage + ppe_available)
                    storage = storage * (storage<awc) * 1.0 +  awc * (storage>=awc) * 1.0
                    storage = 0.0 + ((storage > 0.0) * storage)
                    storage = arcpy.sa.Con(arcpy.sa.IsNull(storage), 0, storage, "Value = 1")
                    storage = arcpy.sa.Con(arcpy.sa.IsNull(awc), awc, storage, "Value = 1")

            # if its the end of the month lets export a snapshot of the storage data
            if dt.day == monthrange(dt.year, dt.month)[1]:
                # last day of month save outputs
                progress_message('Saving Data for %d-%d' % (dt.year, dt.month))
                storage.save(os.path.join(project_folder, 'storage\storage_%d_%02d.tif' % (dt.year, dt.month)))

            if debug_points:
                new_file = not os.path.exists(debug_csv)
                csv_w = csv.DictWriter(open(debug_csv, 'ab'),
                                       fieldnames=['id', 'year', 'month', 'day', 'storage_start','storage','storage_div_awc', 'ppe_month',
                                                   'ppe_day', 'awc', 'ppe_available', 'ppt_month', 'ppt_day',
                                                    'pet_month', 'pet_day'])
                if new_file:
                    csv_w.writeheader()
                for row in arcpy.da.SearchCursor(debug_points, ['FID', 'SHAPE']):
                    coord = " ".join([str(x) for x in row[1]])
                    out_row = {
                        'id': row[0]
                        , 'year': dt.year
                        , 'month': dt.month
                        , 'day': dt.day
                        , 'storage_start' : arcpy.GetCellValue_management(storage_start, coord, "1").getOutput(0)
                        , 'storage': arcpy.GetCellValue_management(storage, coord, "1").getOutput(0)
                        , 'storage_div_awc': arcpy.GetCellValue_management((storage/awc * 1.0), coord, "1").getOutput(0)
                        , 'ppe_month': arcpy.GetCellValue_management(current_month.ppe, coord, "1").getOutput(0)
                        , 'ppe_day': arcpy.GetCellValue_management(current_month.ppe_day, coord, "1").getOutput(0)
                        , 'awc': arcpy.GetCellValue_management(awc, coord, "1").getOutput(0)
                        , 'ppe_available': arcpy.GetCellValue_management(current_month.ppe_day * (storage / awc), coord, "1").getOutput(0)
                        , 'ppt_month': arcpy.GetCellValue_management(current_month.p, coord, "1").getOutput(0)
                        , 'ppt_day': arcpy.GetCellValue_management(current_month.p_day, coord, "1").getOutput(0)
                        , 'pet_month': arcpy.GetCellValue_management(current_month.pet, coord, "1").getOutput(0)
                        , 'pet_day': arcpy.GetCellValue_management(current_month.pet_day, coord, "1").getOutput(0)


                    }
                    csv_w.writerow(out_row)
                del csv_w

            # wrap around if first year
            if (dt.year, dt.month, dt.day) == (start_date.year, 12, 31):
                progress_message("Wrapping First Year")
                if skip_wrap:
                    progress_message("Wrap option wasn't set, skipping wrap.")
                else:
                    # if storage ended full we can skip this.
                    storage_full = arcpy.sa.Abs(storage - awc) <= 0.001
                    storage_full.save(os.path.join(project_folder, 'storage_full.tif'))
                    storage_full_flag = storage_full.minimum == storage_full.maximum == 1
                    del storage_full
                    if not storage_full_flag:
                        wrap = True
                        wrap_start = start_date
                        # save current storage as a year early for restart
                        storage.save(os.path.join(project_folder, 'storage\storage_%d_%02d.tif' % (start_date.year - 1, 12)))

                        # make supporting data for that month if needed
                        pet_path = os.path.join(project_folder, 'pet\pet_%d_%02d.tif' % (start_date.year - 1, 12))
                        if not arcpy.Exists(pet_path):
                            pet = arcpy.Raster(os.path.join(project_folder, 'pet\pet_%d_%02d.tif' % (dt.year, dt.month)))
                            pet.save(pet_path)
                        ppt_path = os.path.join(project_folder, 'ppt\ppt_%d_%02d.tif' % (start_date.year - 1, 12))
                        if not arcpy.Exists(ppt_path):
                            ppt = arcpy.Raster(os.path.join(project_folder, 'ppt\ppt_%d_%02d.tif' % (dt.year, dt.month)))
                            ppt.save(ppt_path)

                        # what is our end month?
                        end_month = False
                        month = 1
                        while not end_month and month < 13:
                            test_month = StorageMonth(project_folder, start_date.year, month)
                            if test_month.ppe.minimum < 0:
                                end_month = True
                            else:
                                month += 1
                        month = month - 1
                        wrap_end = datetime.date(start_date.year, month, monthrange(start_date.year, month)[1])
                        progress_message("Will need to run first year again from %s - %s" % (start_date, wrap_end))

                    else:
                        progress_message("First year ended full, no need to wrap around")

            dt = dt + datetime.timedelta(days=1)
        if wrap and not skip_wrap:
            progress_message("Running first year again")
            # call this process over
            del current_month
            self.run(project_folder, start_date, wrap_end)

        #cleanup
        if arcpy.Exists(os.path.join(project_folder, 'storage_full.tif')):
            arcpy.Delete_management(os.path.join(project_folder, 'storage_full.tif'))
        if arcpy.Exists(os.path.join(project_folder, 'month_full.tif')):
            arcpy.Delete_management(os.path.join(project_folder, 'month_full.tif'))



        return


class StorageDerived(object):
    def __init__(self):
        self.label = "Calculate Storage Derived Rasters"
        self.description = "Calculates AET, Deficit for each month and per annum"
        self.canRunInBackground = False

    def getParameterInfo(self):
        params = list()
        params.append(
            arcpy.Parameter(
                displayName='Project Folder'
                , name='project_folder'
                , datatype='DEFolder'
                , parameterType='Required'
                , direction='Input'
            )
        )
        start = arcpy.Parameter(
            displayName='Start Year'
            , name='start_year'
            , datatype='GPLong'
            , parameterType='Required'
            , direction='Input'
        )
        #start.filter.list = year_filter
        params.append(start)
        end = arcpy.Parameter(
            displayName='End Year'
            , name='end_year'
            , datatype='GPLong'
            , parameterType='Required'
            , direction='Input'
        )
        #end.filter.list = year_filter
        params.append(end)
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        params = {p.name: p.valueAsText for p in parameters}
        self.run(params['project_folder'], int(params['start_year']), int(params['end_year']))
        return

    def run(self, project_folder, start_year, end_year):
        arcpy.env.overwriteOutput = True

        check_folder(os.path.join(project_folder,'storage_delta'))
        check_folder(os.path.join(project_folder,'aet'))
        check_folder(os.path.join(project_folder,'deficit'))
        check_folder(os.path.join(project_folder,'surplus'))
        for year in range(start_year, end_year + 1):
            aet_annual = []
            deficit_annual = []
            surplus_annual = []
            awc = arcpy.Raster(os.path.join(project_folder, 'soils.tif'))
            for month in range(1, 13):
                progress_message("Calculating %s - %s" % (year, month))
                current_date = datetime.date(year, month, 1)
                if month == 1:
                    prev_month = datetime.date(year - 1, 12, 1)
                else:
                    prev_month = datetime.date(year, month - 1, 1)
                storage = arcpy.Raster(
                    os.path.join(project_folder, 'storage\storage_%d_%02d.tif' % (current_date.year, current_date.month)))
                prev_path = os.path.join(project_folder, 'storage\storage_%d_%02d.tif' % (prev_month.year, prev_month.month))
                if arcpy.Exists(prev_path):
                    storage_prev = arcpy.Raster(prev_path)
                elif month==1:
                    storage_prev = arcpy.Raster(os.path.join(project_folder, 'storage\storage_%d_%02d.tif' % (year, 12)))
                else:
                    #should never get here unless someone deletes a file
                    raise WaterBudgetException("Storage not found for %s-%s" %(year,month))
                storage_delta = storage - storage_prev
                storage_delta.save(
                    os.path.join(project_folder, 'storage_delta','storage_delta_%d_%02d.tif' % (current_date.year, current_date.month)))
                ppe = arcpy.Raster(
                    os.path.join(project_folder, 'ppe','ppe_%d_%02d.tif' % (current_date.year, current_date.month)))
                pet = arcpy.Raster(
                    os.path.join(project_folder, 'pet','pet_%d_%02d.tif' % (current_date.year, current_date.month)))
                ppt = arcpy.Raster(
                    os.path.join(project_folder, 'ppt','ppt_%d_%02d.tif' % (current_date.year, current_date.month)))
                aet = 0 + ((ppe < 0) * (ppt - storage_delta)) + ((ppe >= 0) * (pet))
                aet = 0 + (aet <= pet) * aet + (aet > pet) * pet
                aet.save(os.path.join(project_folder, 'aet','aet_%d_%02d.tif' % (current_date.year, current_date.month)))
                deficit = pet - aet
                deficit = 0 + (deficit > 0.0001) * deficit
                deficit.save(
                    os.path.join(project_folder, 'deficit','deficit_%d_%02d.tif' % (current_date.year, current_date.month)))
                surplus = 0 + (arcpy.sa.Abs(storage - awc) < 0.05 ) * (ppt - (aet + storage_delta))
                surplus.save(
                    os.path.join(project_folder, 'surplus','surplus_%d_%02d.tif' % (current_date.year, current_date.month)))
                aet_annual.append(aet)
                deficit_annual.append(deficit)
                surplus_annual.append(surplus)

            progress_message("Calculating %s per annum" % (year))
            annual_aet = aet_annual.pop()
            annual_deficit = deficit_annual.pop()
            annual_surplus = surplus_annual.pop()
            for aet in aet_annual:
                annual_aet = annual_aet + aet
            annual_aet.save(os.path.join(project_folder, 'aet','aet_%d_annual.tif' % (current_date.year)))
            for deficit in deficit_annual:
                annual_deficit = annual_deficit + deficit
            annual_deficit.save(os.path.join(project_folder, 'deficit\deficit_%d_annual.tif' % (current_date.year)))
            for surplus in surplus_annual:
                annual_surplus = annual_surplus + surplus
            annual_surplus.save(os.path.join(project_folder, 'surplus\surplus_%d_annual.tif' % (current_date.year)))

class Checks(object):
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Checks"
        self.description = ""
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Define parameter definitions"""
        params = []
        params.append(
            arcpy.Parameter(
                displayName='Project Folder'
                , name='project_folder'
                , datatype='DEFolder'
                , parameterType='Required'
                , direction='Input'
            )
        )
        start = arcpy.Parameter(
            displayName='Start Year'
            , name='start_year'
            , datatype='GPLong'
            , parameterType='Required'
            , direction='Input'
        )
        #start.filter.list = year_filter
        params.append(start)
        end = arcpy.Parameter(
            displayName='End Year'
            , name='end_year'
            , datatype='GPLong'
            , parameterType='Required'
            , direction='Input'
        )
        #end.filter.list = year_filter
        params.append(end)
        return params

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
        return True

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""
        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""
        return

    def execute(self, parameters, messages):
        """The source code of the tool."""
        params = {p.name: p.valueAsText for p in parameters}
        start_year = int(params['start_year'])
        end_year = int(params['end_year'])
        project_folder = params['project_folder']
        assert end_year >= start_year

        check_folder(os.path.join(project_folder,'check'))

        # perform check rasters
        pet = arcpy.Raster(os.path.join(project_folder, 'pet\pet_%d_annual.tif' % (start_year)))
        aet = arcpy.Raster(os.path.join(project_folder, 'aet\\aet_%d_annual.tif' % (start_year)))
        deficit = arcpy.Raster(os.path.join(project_folder, 'deficit\deficit_%d_annual.tif' % (start_year)))
        ppt = arcpy.Raster(os.path.join(project_folder, 'ppt\ppt_%d_annual.tif' % (start_year)))
        surplus = arcpy.Raster(os.path.join(project_folder, 'surplus\\surplus_%d_annual.tif' % (start_year)))

        if end_year > start_year:
            current_year = start_year+1
            while current_year <= end_year:
                pet = pet + arcpy.Raster(os.path.join(project_folder, 'pet\pet_%d_annual.tif' % (current_year)))
                aet = aet + arcpy.Raster(os.path.join(project_folder, 'aet\\aet_%d_annual.tif' % (current_year)))
                deficit = deficit + arcpy.Raster(os.path.join(project_folder, 'deficit\deficit_%d_annual.tif' % (current_year)))
                ppt = ppt + arcpy.Raster(os.path.join(project_folder, 'ppt\ppt_%d_annual.tif' % (current_year)))
                surplus = arcpy.Raster(os.path.join(project_folder, 'surplus\\surplus_%d_annual.tif' % (current_year)))
                current_year += 1

        pet_check_values = arcpy.sa.Abs(pet - aet - deficit)
        pet_check_values.save(os.path.join(project_folder, 'check\pet_check_value_%d-%d.tif' % (start_year, end_year)))
        pet_check = pet_check_values < 1
        pet_check.save(os.path.join(project_folder, 'check\pet_check_%d-%d.tif' % (start_year, end_year)))

        ppt_check_values = arcpy.sa.Abs(ppt - aet - surplus)
        ppt_check_values.save(os.path.join(project_folder, 'check\ppt_check_value_%d-%d.tif' % (start_year, end_year)))
        ppt_check = ppt_check_values < 1
        ppt_check.save(os.path.join(project_folder, 'check\ppt_check_%d-%d.tif' % (start_year, end_year)))
        return


class Diagnostics(object):
    def __init__(self):
        self.label = "Create Diagnostic File"
        self.description = ""
        self.canRunInBackground = False

    def getParameterInfo(self):
        params = list()
        params.append(
            arcpy.Parameter(
                displayName='Project Folder'
                , name='project_folder'
                , datatype='DEFolder'
                , parameterType='Required'
                , direction='Input'
            )
        )
        debug_points = arcpy.Parameter(
                displayName='Debug Points'
                , name='debug_points'
                , datatype='DEShapefile'
                , parameterType='Required'
                , direction='Input'
            )
        params.append(debug_points)
        id_field = arcpy.Parameter(
            displayName='ID Field'
            , name='id_field'
            , datatype='Field'
            , parameterType='Required'
            , direction='Input'
        )
        id_field.parameterDependencies = [debug_points.name]
        params.append(id_field)
        start = arcpy.Parameter(
            displayName='Start Year'
            , name='start_year'
            , datatype='GPLong'
            , parameterType='Required'
            , direction='Input'
        )
        #start.filter.list = year_filter
        params.append(start)
        end = arcpy.Parameter(
            displayName='End Year'
            , name='end_year'
            , datatype='GPLong'
            , parameterType='Required'
            , direction='Input'
        )
        #end.filter.list = year_filter
        params.append(end)
        outfile = arcpy.Parameter(
            displayName='Outfile'
            , name='outfile'
            , datatype='DEFile'
            , parameterType='Required'
            , direction='Output'
        )
        outfile.filter.filetype = 'csv'
        params.append(outfile)
        variables = arcpy.Parameter(
            displayName='Parameters to export'
            , name='variables'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
            , multiValue=True
        )
        variables.filter.list = ['ppt', 'pet', 'aet', 'storage', 'surplus', 'deficit', 'awc', 'ppe','storage_delta']
        params.append(variables)
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        params = {p.name: p.valueAsText for p in parameters}
        self.run(params['project_folder'], int(params['start_year']), int(params['end_year']), params['outfile'],
                 params['variables'].split(';'), params['debug_points'],params['id_field'])
        return

    def run(self, folder, start, end, outfile, variables, points, id_field='fid'):
        rows = []
        for y in range(start, end + 1):
            for m in range(1, 13):
                for record in arcpy.da.SearchCursor(points, [id_field, 'SHAPE']):
                    coord = " ".join([str(x) for x in record[1]])
                    row = {}
                    row[id_field] = record[0]
                    row['coord'] = record[1]
                    row['year'] = y
                    row['month'] = m
                    for v in variables:
                        if v == 'awc':
                            f = os.path.join(folder, 'soils.tif')
                        else:
                            f = os.path.join(folder, '%s\%s_%d_%02d.tif' % (v, v, y, m))

                        if not arcpy.Exists(f):
                            arcpy.AddError('Failed to find raster ' + f)
                            raise WaterBudgetException('Failed to find raster ' + f)

                        rast = arcpy.Raster(f)
                        rast_value = arcpy.GetCellValue_management(rast, coord, "1").getOutput(0)
                        row[v] = rast_value
                    rows.append(row)
        fieldnames = [id_field, 'coord', 'year', 'month']
        fieldnames.extend(variables)
        csv_w = csv.DictWriter(open(outfile, 'wb'), fieldnames=fieldnames)
        csv_w.writeheader()
        csv_w.writerows(rows)
        del csv_w
